# -*- coding: utf-8 -*-

# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from lifelines import KaplanMeierFitter
from lifelines.statistics import logrank_test
import matplotlib.colors as colors
from .rounding import float_str

def plot_kmf(df,
             condition_col,
             censor_col,
             survival_col,
             threshold=None,
             title=None,
             xlabel=None,
             ylabel=None,
             ax=None,
             with_condition_color="#B38600",
             no_condition_color="#A941AC",
             with_condition_label=None,
             no_condition_label=None,
             ci_show=False,
             print_as_title=False):
    """
    Plot survival curves by splitting the dataset into two groups based on
    condition_col

    if threshold is defined, the groups are split based on being > or <
    condition_col

    if threshold == 'median', the threshold is set to the median of condition_col

    Parameters
    ----------
        df: dataframe
        condition_col: string, column which contains the condition to split on
        survival_col: string, column which contains the survival time
        censor_col: string,
        threshold: int or string, if int, condition_col is thresholded,
                                  if 'median', condition_col thresholded
                                  at its median
        title: Title for the plot, default None
        ax: an existing matplotlib ax, optional, default None
        with_condition_color: str, hex code color for the with-condition curve
        no_condition_color: str, hex code color for the no-condition curve
        print_as_title: bool, optional, whether or not to print text
          within the plot's title vs. stdout, default False
    """
    with_condition_color = colors.hex2color(with_condition_color)
    no_condition_color = colors.hex2color(no_condition_color)
    kmf = KaplanMeierFitter()
    if threshold is not None:
        is_median = threshold == "median"
        if is_median:
            threshold = df[condition_col].median()
        label_suffix = float_str(threshold)
        condition = df[condition_col] > threshold
        default_label_no_condition = "%s ≤ %s" % (condition_col, label_suffix)
        if is_median:
            label_suffix += " (median)"
        default_label_with_condition = "%s > %s" % (condition_col, label_suffix)
    else:
        condition = df[condition_col]
        default_label_with_condition = "= {}".format(condition_col)
        default_label_no_condition = "¬ {}".format(condition_col)

    with_condition_label = with_condition_label or default_label_with_condition
    no_condition_label = no_condition_label or default_label_no_condition

    df_with_condition = df[condition]
    df_no_condition = df[~condition]
    survival_no_condition = df_no_condition[survival_col]
    survival_with_condition = df_with_condition[survival_col]

    event_no_condition = (df_no_condition[censor_col].astype(bool))
    event_with_condition = (df_with_condition[censor_col].astype(bool))

    kmf.fit(survival_no_condition, event_no_condition, label=(no_condition_label))
    if ax:
        kmf.plot(ax=ax, show_censors=True, ci_show=ci_show, color=no_condition_color)
    else:
        ax = kmf.plot(show_censors=True, ci_show=ci_show, color=no_condition_color)

    kmf.fit(survival_with_condition, event_with_condition, label=(with_condition_label))
    plot = kmf.plot(ax=ax, show_censors=True, ci_show=ci_show, color=with_condition_color)

    # Set the y-axis to range 0 to 1
    ax.set_ylim(0, 1)
    y_tick_vals = ax.get_yticks()
    ax.set_yticklabels(["%d" % int(y_tick_val * 100) for y_tick_val in y_tick_vals])

    no_cond_str = "# no condition {}".format(len(survival_no_condition))
    cond_str = "# with condition {}".format(len(survival_with_condition))
    if title:
        ax.set_title(title)
    elif print_as_title:
        ax.set_title("%s | %s" % (no_cond_str, cond_str))
    else:
        print(no_cond_str)
        print(cond_str)

    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)

    results = logrank_test(survival_no_condition,
                           survival_with_condition,
                           event_observed_A=event_no_condition,
                           event_observed_B=event_with_condition)
    results.without_condition_series = survival_no_condition
    results.with_condition_series = survival_with_condition
    return results

def logrank(df,
            condition_col,
            censor_col,
            survival_col,
            threshold=None):
    if threshold is not None:
        if threshold == 'median':
            threshold = df[condition_col].median()
        condition = df[condition_col] > threshold
    else:
        condition = df[condition_col]
    df_with_condition = df[condition]
    df_no_condition = df[~condition]
    survival_no_condition = df_no_condition[survival_col]
    survival_with_condition = df_with_condition[survival_col]
    event_no_condition = (df_no_condition[censor_col].astype(bool))
    event_with_condition = (df_with_condition[censor_col].astype(bool))
    return logrank_test(survival_no_condition,
                        survival_with_condition,
                        event_observed_A=event_no_condition,
                        event_observed_B=event_with_condition)



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from collections import namedtuple

from scipy.stats import mannwhitneyu, fisher_exact
import seaborn as sb
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.metrics import roc_curve
from .model import bootstrap_auc

def vertical_percent(plot, percent=0.1):
    """
    Using the size of the y axis, return a fraction of that size.
    """
    plot_bottom, plot_top = plot.get_ylim()
    return percent * (plot_top - plot_bottom)

def as_numeric(text):
    try:
        return float(text)
    except:
        return None

def hide_ticks(plot, min_tick_value=None, max_tick_value=None):
    """Hide tick values that are outside of [min_tick_value, max_tick_value]"""
    for tick, tick_value in zip(plot.get_yticklabels(), plot.get_yticks()):
        tick_label = as_numeric(tick_value)
        if tick_label:
            if (min_tick_value is not None and tick_label < min_tick_value or 
                 max_tick_value is not None and tick_label > max_tick_value):
                tick.set_visible(False)

def hide_negative_y_ticks(plot):
    hide_ticks(plot, min_tick_value=0)

def only_percentage_ticks(plot):
    """
    Only show ticks from 0.0 to 1.0.
    """
    hide_ticks(plot, min_tick_value=0, max_tick_value=1.0)

def add_significance_indicator(plot, col_a=0, col_b=1, significant=False):
    """
    Add a p-value significance indicator.
    """
    plot_bottom, plot_top = plot.get_ylim()
    # Give the plot a little room for the significance indicator
    line_height = vertical_percent(plot, 0.1)
    # Add some extra spacing below the indicator
    plot_top = plot_top + line_height
    # Add some extra spacing above the indicator
    plot.set_ylim(top=plot_top + line_height * 2)
    color = "black"
    line_top = plot_top + line_height
    plot.plot([col_a, col_a, col_b, col_b], [plot_top, line_top, line_top, plot_top], lw=1.5, color=color)
    indicator = "*" if significant else "ns"
    plot.text((col_a + col_b) * 0.5, line_top, indicator, ha="center", va="bottom", color=color)

def stripboxplot(x, y, data, ax=None, significant=None, **kwargs):
    """
    Overlay a stripplot on top of a boxplot.
    """
    ax = sb.boxplot(
        x=x,
        y=y,
        data=data,
        ax=ax,
        fliersize=0,
        **kwargs
    )

    plot = sb.stripplot(
        x=x,
        y=y,
        data=data,
        ax=ax,
        jitter=kwargs.pop("jitter", 0.05),
        color=kwargs.pop("color", "0.3"),
        **kwargs
    )

    if data[y].min() >= 0:
        hide_negative_y_ticks(plot)
    if significant is not None:
        add_significance_indicator(plot=plot, significant=significant)

    return plot

def sided_str_from_alternative(alternative, condition):
    if alternative is None:
        raise ValueError("Must pick an alternative")
    if alternative == "two-sided":
        return alternative
    # alternative hypothesis: condition is 'less' or 'greater' than no-condition
    op_str = ">" if alternative == "greater" else "<"
    return "one-sided: %s %s not %s" % (condition, op_str, condition)

def get_condition_mask(df, condition, condition_value):
    if condition_value:
        condition_mask = df[condition] == condition_value
    else:
        # This is necessary in the event that condition has a non-bool dtype,
        # such as object. This may happen if a function returns np.nan in
        # addition to True/False (later filtered down to just True/False).
        # ~condition_mask will behave incorrectly if dtype is not bool.
        condition_mask = df[condition].astype("bool")
    return condition_mask

class FishersExactResults(namedtuple("FishersExactResults", ["oddsratio", "p_value", "sided_str", "with_condition1_series", "without_condition1_series", "plot"])):
    def __str__(self):
        return "FishersExactResults(oddsratio=%s, p_value=%s, sided_str='%s')" % (
            self.oddsratio, self.p_value, self.sided_str)

    def __repr__(self):
        return self.__str__()

def fishers_exact_plot(data, condition1, condition2, ax=None,
                       condition1_value=None,
                       alternative="two-sided", **kwargs):
    """
    Perform a Fisher's exact test to compare to binary columns

    Parameters
    ----------
    data: Pandas dataframe
        Dataframe to retrieve information from

    condition1: str
        First binary column to compare (and used for test sidedness)

    condition2: str
        Second binary column to compare

    ax : Axes, default None
        Axes to plot on

    condition1_value:
        If `condition1` is not a binary column, split on =/!= to condition1_value

    alternative:
        Specify the sidedness of the test: "two-sided", "less"
        or "greater"
    """
    plot = sb.barplot(
        x=condition1,
        y=condition2,
        ax=ax,
        data=data,
        **kwargs
    )

    plot.set_ylabel("Percent %s" % condition2)
    condition1_mask = get_condition_mask(data, condition1, condition1_value)
    count_table = pd.crosstab(data[condition1], data[condition2])
    print(count_table)
    oddsratio, p_value = fisher_exact(count_table, alternative=alternative)
    add_significance_indicator(plot=plot, significant=p_value <= 0.05)
    only_percentage_ticks(plot)

    if alternative != "two-sided":
        raise ValueError("We need to better understand the one-sided Fisher's Exact test")
    sided_str = "two-sided"
    print("Fisher's Exact Test: OR: {}, p-value={} ({})".format(oddsratio, p_value, sided_str))
    return FishersExactResults(oddsratio=oddsratio,
                               p_value=p_value,
                               sided_str=sided_str,
                               with_condition1_series=data[condition1_mask][condition2],
                               without_condition1_series=data[~condition1_mask][condition2],
                               plot=plot)

class MannWhitneyResults(namedtuple("MannWhitneyResults", ["U", "p_value", "sided_str", "with_condition_series", "without_condition_series", "plot"])):
    def __str__(self):
        return "MannWhitneyResults(U=%s, p_value=%s, sided_str='%s')" % (
            self.U, self.p_value, self.sided_str)

    def __repr__(self):
        return self.__str__()

def mann_whitney_plot(data,
                      condition,
                      distribution,
                      ax=None,
                      condition_value=None,
                      alternative="two-sided",
                      skip_plot=False,
                      **kwargs):
    """
    Create a box plot comparing a condition and perform a
    Mann Whitney test to compare the distribution in condition A v B

    Parameters
    ----------
    data: Pandas dataframe
        Dataframe to retrieve information from

    condition: str
        Column to use as the splitting criteria

    distribution: str
        Column to use as the Y-axis or distribution in the test

    ax : Axes, default None
        Axes to plot on

    condition_value:
        If `condition` is not a binary column, split on =/!= to condition_value

    alternative:
        Specify the sidedness of the Mann-Whitney test: "two-sided", "less"
        or "greater"

    skip_plot:
        Calculate the test statistic and p-value, but don't plot.
    """
    condition_mask = get_condition_mask(data, condition, condition_value)
    U, p_value = mannwhitneyu(
        data[condition_mask][distribution],
        data[~condition_mask][distribution],
        alternative=alternative
    )

    plot = None
    if not skip_plot:
        plot = stripboxplot(
            x=condition,
            y=distribution,
            data=data,
            ax=ax,
            significant=p_value <= 0.05,
            **kwargs
        )

    sided_str = sided_str_from_alternative(alternative, condition)
    print("Mann-Whitney test: U={}, p-value={} ({})".format(U, p_value, sided_str))
    return MannWhitneyResults(U=U,
                              p_value=p_value,
                              sided_str=sided_str,
                              with_condition_series=data[condition_mask][distribution],
                              without_condition_series=data[~condition_mask][distribution],
                              plot=plot)

class CorrelationResults(namedtuple("CorrelationResults", ["coeff", "p_value", "stat_func", "series_x", "series_y", "plot"])):
    def __str__(self):
        return "CorrelationResults(coeff=%s, p_value=%s, stat_func=%s)" % (
            self.coeff, self.p_value, self.stat_func.__name__)

    def __repr__(self):
        return self.__str__()

def roc_curve_plot(data, value_column, outcome_column, bootstrap_samples=100, ax=None):
    """Create a ROC curve and compute the bootstrap AUC for the given variable and outcome

    Parameters
    ----------
    data : Pandas dataframe
        Dataframe to retrieve information from
    value_column : str
        Column to retrieve the values from
    outcome_column : str
        Column to use as the outcome variable
    bootstrap_samples : int, optional
        Number of bootstrap samples to use to compute the AUC
    ax : Axes, default None
        Axes to plot on

    Returns
    -------
    (mean_bootstrap_auc, roc_plot) : (float, matplotlib plot)
        Mean AUC for the given number of bootstrap samples and the plot
    """
    scores = bootstrap_auc(df=data,
                           col=value_column,
                           pred_col=outcome_column,
                           n_bootstrap=bootstrap_samples)
    mean_bootstrap_auc = scores.mean()
    print("{}, Bootstrap (samples = {}) AUC:{}, std={}".format(
        value_column, bootstrap_samples, mean_bootstrap_auc, scores.std()))

    outcome = data[outcome_column].astype(int)
    values = data[value_column]
    fpr, tpr, thresholds = roc_curve(outcome, values)

    if ax is None:
        ax = plt.gca()

    roc_plot = ax.plot(fpr, tpr, lw=1, label=value_column)

    ax.set_xlim([-0.05, 1.05])
    ax.set_ylim([-0.05, 1.05])
    ax.set_xlabel('False Positive Rate')
    ax.set_ylabel('True Positive Rate')
    ax.legend(loc=2, borderaxespad=0.)
    ax.set_title('{} ROC Curve (n={})'.format(value_column, len(values)))

    return (mean_bootstrap_auc, roc_plot)



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from os import path, makedirs
from shutil import rmtree
import pandas as pd
import seaborn as sb
import json
import warnings
import pprint

# pylint doesn't like this line
# pylint: disable=no-name-in-module
import six.moves.cPickle as pickle
from types import FunctionType

import vap  ## vcf-annotate-polyphen
from sqlalchemy import create_engine

from pyensembl import cached_release

import varcode
from varcode import VariantCollection, EffectCollection, Variant
from mhctools import NetMHCcons, EpitopeCollection
from topiary import predict_epitopes_from_variants, epitopes_to_dataframe
from topiary.sequence_helpers import contains_mutant_residues
from isovar.protein_sequence import variants_to_protein_sequences_dataframe
from pysam import AlignmentFile
from scipy.stats import pearsonr
from collections import defaultdict

from .patient import Patient
from .sample import Sample
from .dataframe_loader import DataFrameLoader
from .utils import first_not_none_param, filter_not_null, InvalidDataError, strip_column_names as _strip_column_names
from .provenance import compare_provenance
from .survival import plot_kmf
from .plot import mann_whitney_plot, fishers_exact_plot, roc_curve_plot, stripboxplot, CorrelationResults
from .collection import Collection
from .varcode_utils import (filter_variants, filter_effects,
                            filter_neoantigens, filter_polyphen)
from .variant_filters import no_filter
from .styling import set_styling
from . import variant_filters

class Cohort(Collection):
    """
    Represents a cohort of `Patient`s.

    Parameters
    __________
    patients : List
        A list of `Patient`s for this cohort.
    cache_dir : str
        Path to store cached results, e.g. cached variant effects.
    kallisto_ensembl_version : int
        Cached release version to use from pyensembl to annotate Kallisto data
    cache_results : bool
        Whether or not to cache results.
    extra_df_loaders : List
        List of `DataFrameLoader`s to include as join options `Cohort`.
    join_with : str or List
        The name of one or more `DataFrameLoader`s to join with by default.
    join_how : str
        What type of default join to use for joining `DataFrameLoader`s.
    filter_fn : Function
        Specify a default filter function for `load_variants`, `load_effects`,
        `functions.missense_snv_count`, etc.
    normalized_per_mb : bool
        Whether or not to normalize by number of loci.
    min_coverage_depth_normal : int
        When counting number of exonic loci, only count loci with at least this much normal depth.
    min_coverage_depth_tumor : int
        When counting number of exonic loci, only count loci with at least this much tumor depth.
    responder_pfs_equals_os : bool
        Ensure that the PFS values for responders (not progressed) are equal to
        OS values.
    check_provenance : bool
        Verify that the cached provenance is equal to the current environment.
    print_provenance : bool
        Print a summary of cache file provenance.
    polyphen_dump_path : str
        Path to a Polyphen database dump.
    pageant_coverage_path : str
        Path to Pageant CoverageDepth output.
    benefit_plot_name : str
        What word to use for "benefit" when plotting.
    variant_type : {"snv", "indel"}, optional
        Load variants of a specific type, default "snv"
    merge_type : {"union", "intersection"}, optional
        Use this method to merge multiple variant sets for a single patient, default "union"
    """
    def __init__(self,
                 patients,
                 cache_dir,
                 kallisto_ensembl_version=None,
                 cache_results=True,
                 extra_df_loaders=[],
                 join_with=None,
                 join_how="inner",
                 filter_fn=None,
                 normalized_per_mb=False,
                 min_coverage_normal_depth=0,
                 min_coverage_tumor_depth=0,
                 responder_pfs_equals_os=False,
                 check_provenance=False,
                 print_provenance=True,
                 polyphen_dump_path=None,
                 pageant_coverage_path=None,
                 benefit_plot_name="Benefit",
                 variant_type="snv",
                 merge_type="union"):
        Collection.__init__(
            self,
            elements=patients)
        # TODO: Patients shouldn't actually need to reference their Cohort; remove
        # this when patient-specific functions all live in Patient.
        for patient in patients:
            patient.cohort = self
        self.cache_dir = cache_dir
        self.cache_results = cache_results
        self.kallisto_ensembl_version = kallisto_ensembl_version

        df_loaders = [
            DataFrameLoader("kallisto", self.load_kallisto),
            DataFrameLoader("cufflinks", self.load_cufflinks),
            DataFrameLoader("ensembl_coverage", self.load_ensembl_coverage)]
        df_loaders.extend(extra_df_loaders)
        self.df_loaders = df_loaders
        self.join_with = join_with
        self.join_how = join_how
        self.filter_fn = filter_fn
        self.normalized_per_mb = normalized_per_mb
        self.min_coverage_normal_depth = min_coverage_normal_depth
        self.min_coverage_tumor_depth = min_coverage_tumor_depth
        self.responder_pfs_equals_os = responder_pfs_equals_os
        self.check_provenance = check_provenance
        self.polyphen_dump_path = polyphen_dump_path
        self.pageant_coverage_path = pageant_coverage_path
        self.benefit_plot_name = benefit_plot_name
        self.variant_type = variant_type
        self.merge_type = merge_type

        assert variant_type in ["snv", "indel"], "Unknown variant type: %s" % variant_type

        self.verify_id_uniqueness()
        self.verify_survival()
        self.dataframe_hash = None

        self.cache_names = {"variant": "cached-variants",
                            "effect": "cached-effects",
                            "nonsynonymous_effect": "cached-nonsynonymous-effects",
                            "neoantigen": "cached-neoantigens",
                            "expressed_neoantigen": "cached-expressed-neoantigens",
                            "polyphen": "cached-polyphen-annotations",
                            "isovar": "cached-isovar-output"}
        if print_provenance:
            pprint.pprint(self.summarize_data_sources())

        set_styling()

    def verify_id_uniqueness(self):
        patient_ids = set([patient.id for patient in self])
        if len(patient_ids) != len(self):
            raise ValueError("Non-unique patient IDs")

    def verify_survival(self):
        cohort_dataframe = self.as_dataframe()
        if not (cohort_dataframe["pfs"] <=
                cohort_dataframe["os"]).all():
            raise InvalidDataError("PFS should be <= OS, but PFS is larger than OS for some patients.")

        def func(row):
            if row["pfs"] < row["os"]:
                if not row["progressed_or_deceased"]:
                    raise InvalidDataError(
                        "A patient did not progress despite PFS being less than OS. "
                        "Full row: %s" % row)
        if self.responder_pfs_equals_os:
            cohort_dataframe.apply(func, axis=1)

    def _as_dataframe_unmodified(self, join_with = None, join_how = None):
        # Use join_with if specified, otherwise fall back to what is defined in the class
        join_with = first_not_none_param([join_with, self.join_with], default=[])
        if type(join_with) == str:
            join_with = [join_with]

        # Convert strings to DataFrameLoader objects
        df_loaders = [df_loader for df_loader in self.df_loaders if df_loader.name in join_with]

        # Use join_how if specified, otherwise fall back to what is defined in the class
        join_how = first_not_none_param([join_how, self.join_how], default="inner")

        patient_rows = []
        for patient in self:
            row = {} if patient.additional_data is None else patient.additional_data.copy()
            row["patient_id"] = patient.id
            for clinical_col in ["benefit", "os", "pfs", "deceased",
                                "progressed", "progressed_or_deceased"]:
                row[clinical_col] = getattr(patient, clinical_col)

            patient_rows.append(row)
        df = pd.DataFrame.from_records(patient_rows)

        # Are any columns duplicated in the DataFrame(s) to be joined?
        # If so, rename those columns to be suffixed by the DataFrameLoader
        # name.
        df_loader_dfs = {}
        col_counts = defaultdict(int)
        for df_loader in df_loaders:
            df_loader_dfs[df_loader] = df_loader.load_dataframe()
            for col in df_loader_dfs[df_loader].columns:
                col_counts[col] += 1
        for col, count in col_counts.items():
            # Don't rename columns that are not duplicated.
            if count > 1:
                for df_loader, loaded_df in df_loader_dfs.items():
                    # Don't rename a column that will be joined on.
                    if col != "patient_id" and col != df_loader.join_on:
                        loaded_df.rename(columns={col: "%s_%s" % (col, df_loader.name)}, inplace=True)

        for df_loader, loaded_df in df_loader_dfs.items():
            old_len_df = len(df)
            df = df.merge(
                loaded_df,
                left_on="patient_id",
                right_on=df_loader.join_on,
                how=join_how)
            print("%s join with %s: %d to %d rows" % (
                join_how,
                df_loader.name,
                old_len_df,
                len(df)))
        self.dataframe_hash = hash(str(df.sort_values("patient_id")))
        return df

    def as_dataframe(self, on=None, col=None, join_with=None, join_how=None,
                     rename_cols=False, keep_paren_contents=True, **kwargs):
        """
        Return this Cohort as a DataFrame, and optionally include additional columns
        using `on`.

        on : function or list or dict or str, optional
            - A function that creates a new column for comparison, e.g. count.snv_count.
            - Or a list of column-generating functions.
            - Or a map of new column names to their column-generating functions.
            - A column name that gets returned with the original dataframe as (col, df).
        col : str, optional
            If `on` is a function generating a column, col is the name of that column.
            If `on` is a str specifying a column, col is the name of a copy of that column.
            If None, defaults to the name of the function.

        If `on` is a function or functions, kwargs is passed to those functions.
        Otherwise kwargs is ignored.

        Other parameters
        ----------------
        `rename_cols`: (bool)
            if True, then return columns using "stripped" column names
            ("stripped" means lower-case names without punctuation other than `_`)
            See `utils.strip_column_names` for more details
            defaults to False
        `keep_paren_contents`: (bool)
            if True, then contents of column names within parens are kept.
            if False, contents of column names within-parens are dropped.
            Defaults to True
        ----------

        Return : tuple or DataFrame
            <dataframe>
            or (<name of new column>, <dataframe with that new column>)
            or (<list of names of new columns>, <dataframe with those new columns>)
        """
        df = self._as_dataframe_unmodified(join_with=join_with, join_how=join_how)
        if on is None:
            return df

        if type(on) == str:
            if col is not None:
                df[col] = df[on]
                return (col, df)
            return (on, df)

        def apply_func(on, col, df):
            """
            Sometimes we have functions that, by necessity, have more parameters
            than just `row`. We construct a function with just the `row` parameter
            so it can be sent to `DataFrame.apply`. We hackishly pass `cohort`
            (as `self`) along if the function accepts a `cohort` argument.
            """
            if col is None:
                # Use the function name, or "column" for lambdas, if no
                # name is provided for the newly created column.
                col = on.__name__ if not is_lambda(on) else "column"
            on_argnames = on.__code__.co_varnames
            if "cohort" not in on_argnames:
                func = lambda row: on(row=row, **kwargs)
            else:
                func = lambda row: on(row=row, cohort=self, **kwargs)
            df[col] = df.apply(func, axis=1)
            return (col, df)

        def is_lambda(func):
            return func.__name__ == (lambda: None).__name__

        if type(on) == FunctionType:
            return apply_func(on, col, df)

        # For multiple functions, don't allow kwargs since we won't know which functions
        # they apply to.
        if len(kwargs) > 0:
            raise ValueError("kwargs are not supported when collecting multiple functions "
                             "as we don't know which function they apply to.")

        if type(on) == dict:
            cols = []
            for key, value in on.iteritems():
                col, df = apply_func(on=value, col=key, df=df)
                cols.append(col)
        if type(on) == list:
            cols = []
            for i, elem in enumerate(on):
                col = elem.__name__ if not is_lambda(elem) else "column_%d" % i
                col, df = apply_func(on=elem, col=col, df=df)
                cols.append(col)

        if rename_cols:
            rename_dict = _strip_column_names(df.columns, keep_paren_contents=keep_paren_contents)
            df.rename(columns=rename_dict, inplace=True)
            cols = [rename_dict[col] for col in cols]
        return (cols, df)

    def load_dataframe(self, df_loader_name):
        """
        Instead of joining a DataFrameJoiner with the Cohort in `as_dataframe`, sometimes
        we may want to just directly load a particular DataFrame.
        """
        # Get the DataFrameLoader object corresponding to this name.
        df_loaders = [df_loader for df_loader in self.df_loaders if df_loader.name == df_loader_name]

        if len(df_loaders) == 0:
            raise ValueError("No DataFrameLoader with name %s" % df_loader_name)
        if len(df_loaders) > 1:
            raise ValueError("Multiple DataFrameLoaders with name %s" % df_loader_name)

        return df_loaders[0].load_dataframe()

    def generate_provenance(self):
        module_names = ["cohorts", "pyensembl", "varcode", "mhctools", "topiary", "isovar", "scipy", "numpy", "pandas"]
        module_versions = [__import__(module_name).__version__ for module_name in module_names]
        return dict(zip(module_names, module_versions))

    def load_provenance(self, patient_cache_dir):
        with open(path.join(patient_cache_dir, "PROVENANCE"), "r") as f:
            return json.load(f)

    def save_provenance(self, patient_cache_dir, provenance):
        with open(path.join(patient_cache_dir, "PROVENANCE"), "w") as f:
            json.dump(provenance, f)

    def load_from_cache(self, cache_name, patient_id, file_name):
        if not self.cache_results:
            return None

        cache_dir = path.join(self.cache_dir, cache_name)
        patient_cache_dir = path.join(cache_dir, str(patient_id))
        cache_file = path.join(patient_cache_dir, file_name)

        if not path.exists(cache_file):
            return None

        if self.check_provenance:
            num_discrepant = compare_provenance(
                this_provenance = self.generate_provenance(), 
                other_provenance = self.load_provenance(patient_cache_dir),
                left_outer_diff = "In current environment but not cached in %s for patient %s" % (cache_name, patient_id),
                right_outer_diff = "In cached %s for patient %s but not current" % (cache_name, patient_id)
                )

        if path.splitext(cache_file)[1] == ".csv":
            return pd.read_csv(cache_file, dtype={"patient_id": object})
        else:
            with open(cache_file, "rb") as f:
                return pickle.load(f)

    def save_to_cache(self, obj, cache_name, patient_id, file_name):
        if not self.cache_results:
            return

        cache_dir = path.join(self.cache_dir, cache_name)
        patient_cache_dir = path.join(cache_dir, str(patient_id))
        cache_file = path.join(patient_cache_dir, file_name)

        if not path.exists(patient_cache_dir):
            makedirs(patient_cache_dir)

        if type(obj) == pd.DataFrame:
            obj.to_csv(cache_file, index=False)
        else:
            with open(cache_file, "wb") as f:
                pickle.dump(obj, f)

        provenance = self.generate_provenance()
        self.save_provenance(patient_cache_dir, provenance)

    def iter_patients(self, patients):
        if patients is None:
            return self
        return iter(patients)

    def patient_from_id(self, id):
        for patient in self:
            if patient.id == id:
                return patient
        raise ValueError("No patient with ID %s found" % id)

    def load_variants(self, patients=None, filter_fn=None):
        """Load a dictionary of patient_id to varcode.VariantCollection

        Parameters
        ----------
        patients : str, optional
            Filter to a subset of patients
        filter_fn : function
            Takes a FilterableVariant and returns a boolean. Only variants returning True are preserved.
            Overrides default self.filter_fn. `None` passes through to self.filter_fn.

        Returns
        -------
        merged_variants
            Dictionary of patient_id to VariantCollection
        """
        filter_fn = first_not_none_param([filter_fn, self.filter_fn], no_filter)
        patient_variants = {}

        for patient in self.iter_patients(patients):
            variants = self._load_single_patient_variants(patient, filter_fn)
            if variants is not None:
                patient_variants[patient.id] = variants
        return patient_variants

    def _load_single_patient_variants(self, patient, filter_fn):
        failed_io = False
        try:
            cached_file_name = "%s-%s-variants.pkl" % (self.variant_type, self.merge_type)
            cached = self.load_from_cache(self.cache_names["variant"], patient.id, cached_file_name)
            if cached is not None:
                return filter_variants(variant_collection=cached,
                                       patient=patient,
                                       filter_fn=filter_fn)
            vcf_paths = patient.snv_vcf_paths if self.variant_type == "snv" else patient.indel_vcf_paths
            variant_collections = [
                (vcf_path, varcode.load_vcf_fast(vcf_path))
                for vcf_path in vcf_paths
            ]
        except IOError:
            failed_io = True

        # Note that this is the number of variant collections and not the number of
        # variants. 0 variants will lead to 0 neoantigens, for example, but 0 variant
        # collections will lead to NaN variants and neoantigens.
        if failed_io or len(variant_collections) == 0:
            print("Variants did not exist for patient %s" % patient.id)
            return None

        if len(variant_collections) == 1:
            # There is nothing to merge
            vcf_path, variants = variant_collections[0]
            merged_variants = variants
        else:
            merged_variants = self._merge_variant_collections(dict(variant_collections), self.merge_type)

        self.save_to_cache(merged_variants, self.cache_names["variant"], patient.id, cached_file_name)
        return filter_variants(variant_collection=merged_variants,
                               patient=patient,
                               filter_fn=filter_fn)

    def _merge_variant_collections(self, vcf_to_variant_collections, merge_type):
        assert merge_type in ["union", "intersection"], "Unknown merge type: %s" % merge_type
        combined_variants = [set(vc.elements) for (path, vc) in vcf_to_variant_collections.items()]
        if merge_type == "union":
            merged_variants = VariantCollection(set.union(*combined_variants))
        elif merge_type == "intersection":
            merged_variants = VariantCollection(set.intersection(*combined_variants))

        for variant in merged_variants:
            metadata = [
                (vcf_path, vc.metadata[variant])
                for (vcf_path, vc) in vcf_to_variant_collections.items()
                if variant in vc.metadata
            ]

            merged_variants.metadata[variant] = dict(metadata)

        return merged_variants

    def load_polyphen_annotations(self, as_dataframe=False,
                                  filter_fn=None):
        """Load a dataframe containing polyphen2 annotations for all variants

        Parameters
        ----------
        database_file : string, sqlite
            Path to the WHESS/Polyphen2 SQLite database.
            Can be downloaded and bunzip2"ed from http://bit.ly/208mlIU
        filter_fn : function
            Takes a FilterablePolyphen and returns a boolean.
            Only annotations returning True are preserved.
            Overrides default self.filter_fn. `None` passes through to self.filter_fn.

        Returns
        -------
        annotations
            Dictionary of patient_id to a DataFrame that contains annotations
        """
        filter_fn = first_not_none_param([filter_fn, self.filter_fn], no_filter)
        patient_annotations = {}
        for patient in self:
            annotations = self._load_single_patient_polyphen(
                patient,
                filter_fn=filter_fn)
            if annotations is not None:
                annotations["patient_id"] = patient.id
                patient_annotations[patient.id] = annotations
        if as_dataframe:
            return pd.concat(patient_annotations.values())
        return patient_annotations

    def _load_single_patient_polyphen(self, patient, filter_fn):
        cache_name = self.cache_names["polyphen"]
        cached_file_name = "polyphen-annotations.csv"

        # Don't filter here, as these variants are used to generate the
        # PolyPhen cache; and cached items are never filtered.
        variants = self._load_single_patient_variants(patient,
                                                      filter_fn=None)
        if variants is None:
            return None

        cached = self.load_from_cache(cache_name, patient.id, cached_file_name)
        if cached is not None:
            return filter_polyphen(polyphen_df=cached,
                                   variant_collection=variants,
                                   patient=patient,
                                   filter_fn=filter_fn)

        engine = create_engine("sqlite:///{}".format(self.polyphen_dump_path))
        conn = engine.connect()

        df = pd.DataFrame(columns=["chrom", "pos", "ref", "alt",
                                   "annotation_found", "gene", "protein",
                                   "aa_change", "hvar_pred", "hvar_prob",
                                   "hdiv_pred", "hdiv_prob"])
        for variant in variants:
            chrom = "chr{}".format(getattr(variant, "contig", None))
            pos = getattr(variant, "start", None)
            ref = getattr(variant, "ref", None)
            alt = getattr(variant, "alt",  None)
            annotation = vap.annotate_variant(conn, chrom, pos, ref, alt)
            datum = {"chrom": chrom,
                     "pos": pos,
                     "ref": ref,
                     "alt": alt,
                     "annotation_found": annotation is not None}
            attributes = ["gene", "protein", "aa_change",
                          "hvar_pred", "hvar_prob",
                          "hdiv_pred", "hdiv_prob"]
            for attr in attributes:
                datum[attr] = getattr(annotation, attr, None)
            df = df.append(datum, ignore_index=True)
        df["pos"] = df["pos"].astype("int")
        df["annotation_found"] = df["annotation_found"].astype("bool")
        self.save_to_cache(df, cache_name, patient.id, cached_file_name)
        return filter_polyphen(polyphen_df=df,
                               variant_collection=variants,
                               patient=patient,
                               filter_fn=filter_fn)

    def load_effects(self, patients=None, only_nonsynonymous=False,
                     filter_fn=None):
        """Load a dictionary of patient_id to varcode.EffectCollection

        Note that this only loads one effect per variant.

        Parameters
        ----------
        patients : str, optional
            Filter to a subset of patients
        only_nonsynonymous : bool, optional
            If true, load only nonsynonymous effects, default False
        filter_fn : function
            Takes a FilterableEffect and returns a boolean. Only effects returning True are preserved.
            Overrides default self.filter_fn. `None` passes through to self.filter_fn.

        Returns
        -------
        effects
             Dictionary of patient_id to varcode.EffectCollection
        """
        filter_fn = first_not_none_param([filter_fn, self.filter_fn], no_filter)
        patient_effects = {}
        for patient in self.iter_patients(patients):
            effects = self._load_single_patient_effects(
                patient, only_nonsynonymous, filter_fn)
            if effects is not None:
                patient_effects[patient.id] = effects
        return patient_effects

    def _load_single_patient_effects(self, patient, only_nonsynonymous, filter_fn):
        cached_file_name = "%s-%s-effects.pkl" % (self.variant_type, self.merge_type)

        # Don't filter here, as these variants are used to generate the
        # effects cache; and cached items are never filtered.
        variants = self._load_single_patient_variants(patient, filter_fn=None)
        if variants is None:
            return None

        if only_nonsynonymous:
            cached = self.load_from_cache(self.cache_names["nonsynonymous_effect"], patient.id, cached_file_name)
        else:
            cached = self.load_from_cache(self.cache_names["effect"], patient.id, cached_file_name)
        if cached is not None:
            return filter_effects(effect_collection=cached,
                                  variant_collection=variants,
                                  patient=patient,
                                  filter_fn=filter_fn)

        effects = variants.effects()

        # Always take the top priority effect per variant so we end up with a single
        # effect per variant.
        nonsynonymous_effects = EffectCollection(
            effects.drop_silent_and_noncoding().top_priority_effect_per_variant().values())
        effects = EffectCollection(effects.top_priority_effect_per_variant().values())

        self.save_to_cache(effects, self.cache_names["effect"], patient.id, cached_file_name)
        self.save_to_cache(nonsynonymous_effects, self.cache_names["nonsynonymous_effect"], patient.id, cached_file_name)
        return filter_effects(
            effect_collection=(
                nonsynonymous_effects if only_nonsynonymous else effects),
            variant_collection=variants,
            patient=patient,
            filter_fn=filter_fn)

    def load_kallisto(self):
        """
        Load Kallisto transcript quantification data for a cohort

        Parameters
        ----------

        Returns
        -------
        kallisto_data : Pandas dataframe
            Pandas dataframe with Kallisto data for all patients
            columns include patient_id, gene_name, est_counts
        """
        kallisto_data = pd.concat(
            [self._load_single_patient_kallisto(patient) for patient in self],
            copy=False
        )

        if self.kallisto_ensembl_version is None:
            raise ValueError("Required a kallisto_ensembl_version but none was specified")

        ensembl_release = cached_release(self.kallisto_ensembl_version)

        kallisto_data['gene_name'] = \
            kallisto_data['target_id'].map(lambda t: ensembl_release.gene_name_of_transcript_id(t))

        # sum counts across genes
        kallisto_data = \
            kallisto_data.groupby(['patient_id', 'gene_name'])[['est_counts']].sum().reset_index()

        return kallisto_data

    def _load_single_patient_kallisto(self, patient):
        """
        Load Kallisto gene quantification given a patient

        Parameters
        ----------
        patient : Patient

        Returns
        -------
        data: Pandas dataframe
            Pandas dataframe of sample's Kallisto data
            columns include patient_id, target_id, length, eff_length, est_counts, tpm
        """
        data = pd.read_csv(patient.tumor_sample.kallisto_path, sep="\t")
        data["patient_id"] = patient.id
        return data

    def load_cufflinks(self, filter_ok=True):
        """
        Load a Cufflinks gene expression data for a cohort

        Parameters
        ----------
        filter_ok : bool, optional
            If true, filter Cufflinks data to row with FPKM_status == "OK"

        Returns
        -------
        cufflinks_data : Pandas dataframe
            Pandas dataframe with Cufflinks data for all patients
            columns include patient_id, gene_id, gene_short_name, FPKM, FPKM_conf_lo, FPKM_conf_hi
        """
        return \
            pd.concat(
                [self._load_single_patient_cufflinks(patient, filter_ok) for patient in self],
                copy=False
        )

    def _load_single_patient_cufflinks(self, patient, filter_ok):
        """
        Load Cufflinks gene quantification given a patient

        Parameters
        ----------
        patient : Patient
        filter_ok : bool, optional
            If true, filter Cufflinks data to row with FPKM_status == "OK"

        Returns
        -------
        data: Pandas dataframe
            Pandas dataframe of sample's Cufflinks data
            columns include patient_id, gene_id, gene_short_name, FPKM, FPKM_conf_lo, FPKM_conf_hi
        """
        data = pd.read_csv(patient.tumor_sample.cufflinks_path, sep="\t")
        data["patient_id"] = patient.id

        if filter_ok:
            # Filter to OK FPKM counts
            data = data[data["FPKM_status"] == "OK"]
        return data

    def load_neoantigens(self, patients=None, only_expressed=False,
                         epitope_lengths=[8, 9, 10, 11], ic50_cutoff=500,
                         process_limit=10, max_file_records=None,
                         filter_fn=None):
        filter_fn = first_not_none_param([filter_fn, self.filter_fn], no_filter)

        dfs = {}
        for patient in self.iter_patients(patients):
            df_epitopes = self._load_single_patient_neoantigens(
                patient=patient,
                only_expressed=only_expressed,
                epitope_lengths=epitope_lengths,
                ic50_cutoff=ic50_cutoff,
                process_limit=process_limit,
                max_file_records=max_file_records,
                filter_fn=filter_fn)
            if df_epitopes is not None:
                dfs[patient.id] = df_epitopes
        return dfs

    def _load_single_patient_neoantigens(self, patient, only_expressed, epitope_lengths,
                                         ic50_cutoff, process_limit, max_file_records,
                                         filter_fn):
        cached_file_name = "%s-%s-neoantigens.csv" % (self.variant_type, self.merge_type)

        # Don't filter here, as these variants are used to generate the
        # neoantigen cache; and cached items are never filtered.
        variants = self._load_single_patient_variants(patient, filter_fn=None)
        if variants is None:
            return None

        if patient.hla_alleles is None:
            print("HLA alleles did not exist for patient %s" % patient.id)
            return None

        if only_expressed:
            cached = self.load_from_cache(self.cache_names["expressed_neoantigen"], patient.id, cached_file_name)
        else:
            cached = self.load_from_cache(self.cache_names["neoantigen"], patient.id, cached_file_name)
        if cached is not None:
            return filter_neoantigens(neoantigens_df=cached,
                                      variant_collection=variants,
                                      patient=patient,
                                      filter_fn=filter_fn)

        mhc_model = NetMHCcons(
            alleles=patient.hla_alleles,
            epitope_lengths=epitope_lengths,
            max_file_records=max_file_records,
            process_limit=process_limit)
        if only_expressed:
            df_isovar = self.load_single_patient_isovar(patient=patient,
                                                         variants=variants,
                                                         epitope_lengths=epitope_lengths)

            # Map from isovar rows to protein sequences
            isovar_rows_to_protein_sequences = dict([
                (frozenset(row.to_dict().items()), row["amino_acids"]) for (i, row) in df_isovar.iterrows()])

            # MHC binding prediction
            epitopes = mhc_model.predict(isovar_rows_to_protein_sequences)

            # Call `get_filtered_isovar_epitopes` in order to only include peptides that
            # overlap a variant; without this filter, when we use
            # protein_sequence_length above, some 8mers generated from a 21mer source will
            # not overlap a variant.
            df_epitopes = self.get_filtered_isovar_epitopes(
                epitopes, ic50_cutoff=ic50_cutoff).dataframe()
            # Store chr/pos/ref/alt in the cached DataFrame so we can filter based on
            # the variant later.
            for variant_column in ["chr", "pos", "ref", "alt"]:
                # Be consistent with Topiary's output of "start" rather than "pos".
                # Isovar, on the other hand, outputs "pos".
                # See https://github.com/hammerlab/topiary/blob/5c12bab3d47bd86d396b079294aff141265f8b41/topiary/converters.py#L50
                df_column = "start" if variant_column == "pos" else variant_column
                df_epitopes[df_column] = df_epitopes.source_sequence_key.apply(
                    lambda key: dict(key)[variant_column])
            df_epitopes["patient_id"] = patient.id

            self.save_to_cache(df_epitopes, self.cache_names["expressed_neoantigen"], patient.id, cached_file_name)
        else:
            epitopes = predict_epitopes_from_variants(
                variants=variants,
                mhc_model=mhc_model,
                ic50_cutoff=ic50_cutoff,
                # Only include peptides with a variant
                only_novel_epitopes=True)
            df_epitopes = epitopes_to_dataframe(epitopes)
            df_epitopes["patient_id"] = patient.id

            self.save_to_cache(df_epitopes, self.cache_names["neoantigen"], patient.id, cached_file_name)

        return filter_neoantigens(neoantigens_df=df_epitopes,
                                  variant_collection=variants,
                                  patient=patient,
                                  filter_fn=filter_fn)

    def get_filtered_isovar_epitopes(self, epitopes, ic50_cutoff):
        """
        Mostly replicates topiary.build_epitope_collection_from_binding_predictions

        Note: topiary needs to do fancy stuff like subsequence_protein_offset + binding_prediction.offset
        in order to figure out whether a variant is in the peptide because it only has the variant's
        offset into the full protein; but isovar gives us the variant's offset into the protein subsequence
        (dictated by protein_sequence_length); so all we need to do is map that onto the smaller 8-11mer
        peptides generated by mhctools.
        """
        mutant_binding_predictions = []
        for binding_prediction in epitopes:
            peptide = binding_prediction.peptide
            peptide_offset = binding_prediction.offset
            isovar_row = dict(binding_prediction.source_sequence_key)
            is_mutant = contains_mutant_residues(
                peptide_start_in_protein=peptide_offset,
                peptide_length=len(peptide),
                mutation_start_in_protein=isovar_row["variant_aa_interval_start"],
                mutation_end_in_protein=isovar_row["variant_aa_interval_end"])
            if is_mutant and binding_prediction.value <= ic50_cutoff:
                mutant_binding_predictions.append(binding_prediction)
        return EpitopeCollection(mutant_binding_predictions)

    def load_single_patient_isovar(self, patient, variants, epitope_lengths):
        # TODO: different epitope lengths, and other parameters, should result in
        # different caches
        isovar_cached_file_name = "%s-%s-isovar.csv" % (self.variant_type, self.merge_type)
        df_isovar = self.load_from_cache(self.cache_names["isovar"], patient.id, isovar_cached_file_name)
        if df_isovar is not None:
            return df_isovar

        import logging
        logging.disable(logging.INFO)
        if patient.tumor_sample is None:
            raise ValueError("Patient %s has no tumor sample" % patient.id)
        if patient.tumor_sample.bam_path_rna is None:
            raise ValueError("Patient %s has no tumor RNA BAM path" % patient.id)
        rna_bam_file = AlignmentFile(patient.tumor_sample.bam_path_rna)
        from isovar.default_parameters import (
            MIN_TRANSCRIPT_PREFIX_LENGTH,
            MAX_REFERENCE_TRANSCRIPT_MISMATCHES
        )

        # To ensure that e.g. 8-11mers overlap substitutions, we need at least this
        # sequence length: (max peptide length * 2) - 1
        # Example:
        # 123456789AB
        #           123456789AB
        # AAAAAAAAAAVAAAAAAAAAA
        protein_sequence_length = (max(epitope_lengths) * 2) - 1
        df_isovar = variants_to_protein_sequences_dataframe(
            variants=variants,
            samfile=rna_bam_file,
            protein_sequence_length=protein_sequence_length,
            min_reads_supporting_rna_sequence=3, # Per Alex R.'s suggestion
            min_transcript_prefix_length=MIN_TRANSCRIPT_PREFIX_LENGTH,
            max_transcript_mismatches=MAX_REFERENCE_TRANSCRIPT_MISMATCHES,
            max_protein_sequences_per_variant=1, # Otherwise we might have too much neoepitope diversity
            min_mapping_quality=1)
        self.save_to_cache(df_isovar, self.cache_names["isovar"], patient.id, isovar_cached_file_name)
        return df_isovar

    def load_ensembl_coverage(self):
        if self.pageant_coverage_path is None:
            raise ValueError("Need a Pageant CoverageDepth path to load ensembl coverage values")
        return variant_filters.load_ensembl_coverage(
            cohort=self,
            coverage_path=self.pageant_coverage_path,
            min_normal_depth=self.min_coverage_normal_depth,
            min_tumor_depth=self.min_coverage_tumor_depth)

    def clear_caches(self):
        for cache in self.cache_names.keys():
            self.clear_cache(cache)

    def clear_cache(self, cache):
        cache_path = path.join(self.cache_dir, self.cache_names[cache])
        if path.exists(cache_path):
            rmtree(cache_path)

    def cohort_columns(self):
        cohort_dataframe = self.as_dataframe()
        column_types = [cohort_dataframe[col].dtype for col in cohort_dataframe.columns]
        return dict(zip(list(cohort_dataframe.columns), column_types))

    def plot_roc_curve(self, on, bootstrap_samples=100, col=None, ax=None, **kwargs):
        """Plot an ROC curve for benefit and a given variable

        Parameters
        ----------
        on : str or function
            See `cohort.load.as_dataframe`
        bootstrap_samples : int, optional
            Number of boostrap samples to use to compute the AUC
        ax : Axes, default None
            Axes to plot on

        Returns
        -------
        (mean_auc_score, plot): (float, matplotlib plot)
            Returns the average AUC for the given predictor over `bootstrap_samples`
            and the associated ROC curve
        """
        plot_col, df = self.as_dataframe(on, col=col, **kwargs)
        df = filter_not_null(df, "benefit")
        df = filter_not_null(df, plot_col)
        df.benefit = df.benefit.astype(bool)
        return roc_curve_plot(df, plot_col, "benefit", bootstrap_samples, ax=ax)

    def plot_benefit(self, on, col=None, benefit_col="benefit", label="Response", ax=None,
                     alternative="two-sided", boolean_value_map={},
                     order=None, **kwargs):
        """Plot a comparison of benefit/response in the cohort on a given variable
        """
        no_benefit_plot_name = "No %s" % self.benefit_plot_name
        boolean_value_map = boolean_value_map or {True: self.benefit_plot_name, False: no_benefit_plot_name}
        order = order or [no_benefit_plot_name, self.benefit_plot_name]

        return self.plot_boolean(on=on,
                                 boolean_col=benefit_col,
                                 col=col,
                                 alternative=alternative,
                                 boolean_label=label,
                                 boolean_value_map=boolean_value_map,
                                 order=order,
                                 ax=ax,
                                 **kwargs)

    def plot_boolean(self,
                     on,
                     boolean_col,
                     plot_col=None,
                     boolean_label=None,
                     boolean_value_map={},
                     col=None,
                     order=None,
                     ax=None,
                     alternative="two-sided",
                     **kwargs):
        """Plot a comparison of `boolean_col` in the cohort on a given variable via
        `on` or `col`.

        If the variable (through `on` or `col`) is binary this will compare
        odds-ratios and perform a Fisher's exact test.

        If the variable is numeric, this will compare the distributions through
        a Mann-Whitney test and plot the distributions with box-strip plot

        Parameters
        ----------
        on : str or function or list or dict
            See `cohort.load.as_dataframe`
        plot_col : str, optional
            If on has many columns, this is the one whose values we are plotting.
            If on has a single column, this is unnecessary.
        boolean_col : str
            Column name of boolean column to plot or compare against.
        boolean_label : None, optional
            Label to give boolean column in the plot
        boolean_value_map : dict, optional
            Map of conversions for values in the boolean column, i.e. {True: 'High', False: 'Low'}
        col : str, optional
            If specified, store the result of `on`. See `cohort.load.as_dataframe`
        order : None, optional
            Order of the labels on the x-axis
        ax : None, optional
            Axes to plot on
        alternative : str, optional
            Choose the sidedness of the mannwhitneyu or Fisher's Exact test.

        Returns
        -------
        (Test statistic, p-value): (float, float)

        """
        cols, df = self.as_dataframe(on, col, **kwargs)
        if type(cols) == str:
            if plot_col is not None:
                raise ValueError("plot_col is specified when it isn't ndeeded.")
            plot_col = cols
        elif type(cols) == list:
            if plot_col is None:
                raise ValueError("plot_col must be specified when multiple `on`s are present.")

        df = filter_not_null(df, boolean_col)
        df = filter_not_null(df, plot_col)

        if boolean_label:
            df[boolean_label] = df[boolean_col]
            boolean_col = boolean_label

        condition_value = None
        if boolean_value_map:
            assert set(boolean_value_map.keys()) == set([True, False]), \
                "Improper mapping of boolean column provided"
            df[boolean_col] = df[boolean_col].map(lambda v: boolean_value_map[v])
            condition_value = boolean_value_map[True]

        if df[plot_col].dtype == "bool":
            results = fishers_exact_plot(
                data=df,
                condition1=boolean_col,
                condition2=plot_col,
                condition1_value=condition_value,
                alternative=alternative,
                order=order,
                ax=ax)
        else:
            results = mann_whitney_plot(
                data=df,
                condition=boolean_col,
                distribution=plot_col,
                condition_value=condition_value,
                alternative=alternative,
                order=order,
                ax=ax)
        return results

    def plot_survival(self, 
                      on, 
                      col=None, 
                      how="os", 
                      survival_units="Days", 
                      ax=None, 
                      ci_show=False,
                      with_condition_color="#B38600",
                      no_condition_color="#A941AC",
                      with_condition_label=None,
                      no_condition_label=None,
                      threshold=None, **kwargs):
        """Plot a Kaplan Meier survival curve by splitting the cohort into two groups
        Parameters
        ----------
        on : str or function
            See `cohort.load.as_dataframe`
        col : str, optional
            If specified, store the result of `on`. See `cohort.load.as_dataframe`
        how : {"os", "pfs"}, optional
            Whether to plot OS (overall survival) or PFS (progression free survival)
        survival_units : str
            Unit of time for the survival measure, i.e. Days or Months
        ci_show : bool
            Display the confidence interval around the survival curve
        threshold : int or "median", optional
            Threshold of `col` on which to split the cohort
        """
        assert how in ["os", "pfs"], "Invalid choice of survival plot type %s" % how
        plot_col, df = self.as_dataframe(on, col, **kwargs)
        df = filter_not_null(df, plot_col)
        if df[plot_col].dtype == "bool":
            default_threshold = None
        else:
            default_threshold = "median"
        results = plot_kmf(
            df=df,
            condition_col=plot_col,
            xlabel=survival_units,
            ylabel="Overall Survival (%)" if how == "os" else "Progression-Free Survival (%)",
            censor_col="deceased" if how == "os" else "progressed_or_deceased",
            survival_col=how,
            threshold=threshold if threshold is not None else default_threshold,
            ax=ax,
            ci_show=ci_show,
            with_condition_color=with_condition_color,
            no_condition_color=no_condition_color,
            with_condition_label=with_condition_label,
            no_condition_label=no_condition_label
        )
        return results

    def plot_joint(self, *args, **kwargs):
        warnings.warn("plot_joint is deprecated; use plot_correlation.", Warning)
        return self.plot_correlation(*args, **kwargs)

    def plot_correlation(self, on, on_two=None, x_col=None, plot_type="jointplot", stat_func=pearsonr, show_stat_func=True, plot_kwargs={}, **kwargs):
        """Plot the correlation between two variables.

        Parameters
        ----------
        on : function or list or dict of functions
            See `cohort.load.as_dataframe`
        on_two : function, optional
            Can specify the second function here rather than creating a list.
        x_col : str, optional
            If `on` is a dict, this guarantees we have the expected ordering.
        plot_type : str, optional
            Specify "jointplot" or "boxplot".
        stat_func : function, optional.
            Specify which function to use for the statistical test.
        show_stat_func : bool, optional
            Whether or not to show the stat_func result in the plot itself.
        plot_kwargs : dict, optional
            kwargs to pass through to plotting functions.
        """
        if plot_type not in ["boxplot", "barplot", "jointplot"]:
            raise ValueError("Invalid plot_type %s" % plot_type)
        if on_two is not None:
            on = [on, on_two]
        plot_cols, df = self.as_dataframe(on, **kwargs)
        for plot_col in plot_cols:
            df = filter_not_null(df, plot_col)
        if x_col is None:
            x_col = plot_cols[0]
            y_col = plot_cols[1]
        else:
            if x_col == plot_cols[0]:
                y_col = plot_cols[1]
            else:
                y_col = plot_cols[0]
        series_x = df[x_col]
        series_y = df[y_col]
        coeff, p_value = stat_func(series_x, series_y)
        if plot_type == "jointplot":
            plot = sb.jointplot(data=df, x=x_col, y=y_col,
                                stat_func=stat_func if show_stat_func else None,
                                **plot_kwargs)
        elif plot_type == "boxplot":
            plot = stripboxplot(data=df, x=x_col, y=y_col, **plot_kwargs)
        else:
            plot = sb.barplot(data=df, x=x_col, y=y_col, **plot_kwargs)
        return CorrelationResults(coeff=coeff, p_value=p_value, stat_func=stat_func,
                                  series_x=series_x, series_y=series_y, plot=plot)

    def _list_patient_ids(self):
        """ Utility function to return a list of patient ids in the Cohort
        """
        results = []
        for patient in self:
            results.append(patient.id)
        return(results)

    def summarize_provenance_per_cache(self):
        """Utility function to summarize provenance files for cached items used by a Cohort,
        for each cache_dir that exists. Only existing cache_dirs are summarized.

        This is a summary of provenance files because the function checks to see whether all
        patients data have the same provenance within the cache dir. The function assumes
        that it will be desireable to have all patients data generated using the same
        environment, for each cache type.

        At the moment, most PROVENANCE files contain details about packages used to generat
        e the cached data file. However, this function is generic & so it summarizes the
        contents of those files irrespective of their contents.

        Returns
        ----------
        Dict containing summarized provenance for each existing cache_dir, after checking
        to see that provenance files are identical among all patients in the data frame for
        that cache_dir.

        If conflicting PROVENANCE files are discovered within a cache-dir:
         - a warning is generated, describing the conflict
         - and, a value of `None` is returned in the dictionary for that cache-dir

        See also
        -----------
        * `?cohorts.Cohort.summarize_provenance` which summarizes provenance files among
        cache_dirs.
        * `?cohorts.Cohort.summarize_dataframe` which hashes/summarizes contents of the data
        frame for this cohort.
        """
        provenance_summary = {}
        df = self.as_dataframe()
        for cache in self.cache_names:
            cache_name = self.cache_names[cache]
            cache_provenance = None
            num_discrepant = 0
            this_cache_dir = path.join(self.cache_dir, cache_name)
            if path.exists(this_cache_dir):
                for patient_id in self._list_patient_ids():
                    patient_cache_dir = path.join(this_cache_dir, patient_id)
                    try:
                        this_provenance = self.load_provenance(patient_cache_dir = patient_cache_dir)
                    except:
                        this_provenance = None
                    if this_provenance:
                        if not(cache_provenance):
                            cache_provenance = this_provenance
                        else:
                            num_discrepant += compare_provenance(this_provenance, cache_provenance)
                if num_discrepant == 0:
                    provenance_summary[cache_name] = cache_provenance
                else:
                    provenance_summary[cache_name] = None
        return(provenance_summary)

    def summarize_dataframe(self):
        """Summarize default dataframe for this cohort using a hash function.
        Useful for confirming the version of data used in various reports, e.g. ipynbs
        """
        if self.dataframe_hash:
            return(self.dataframe_hash)
        else:
            df = self._as_dataframe_unmodified()
            return(self.dataframe_hash)

    def summarize_provenance(self):
        """Utility function to summarize provenance files for cached items used by a Cohort.

        At the moment, most PROVENANCE files contain details about packages used to
        generate files. However, this function is generic & so it summarizes the contents
        of those files irrespective of their contents.

        Returns
        ----------
        Dict containing summary of provenance items, among all cache dirs used by the Cohort.

        IE if all provenances are identical across all cache dirs, then a single set of
        provenances is returned. Otherwise, if all provenances are not identical, the provenance
        items per cache_dir are returned.

        See also
        ----------
        `?cohorts.Cohort.summarize_provenance_per_cache` which is used to summarize provenance
        for each existing cache_dir.
        """
        provenance_per_cache = self.summarize_provenance_per_cache()
        summary_provenance = None
        num_discrepant = 0
        for cache in provenance_per_cache:
            if not(summary_provenance):
                ## pick arbitrary provenance & call this the "summary" (for now)
                summary_provenance = provenance_per_cache[cache]
                summary_provenance_name = cache
            ## for each cache, check equivalence with summary_provenance
            num_discrepant += compare_provenance(
                provenance_per_cache[cache],
                summary_provenance,
                left_outer_diff = "In %s but not in %s" % (cache, summary_provenance_name),
                right_outer_diff = "In %s but not in %s" % (summary_provenance_name, cache)
            )
        ## compare provenance across cached items
        if num_discrepant == 0:
            prov = summary_provenance ## report summary provenance if exists
        else:
            prov = provenance_per_cache ## otherwise, return provenance per cache
        return(prov)

    def summarize_data_sources(self):
        """Utility function to summarize data source status for this Cohort, useful for confirming
        the state of data used for an analysis

        Returns
        ----------
        Dictionary with summary of data sources

        Currently contains
        - dataframe_hash: hash of the dataframe (see `?cohorts.Cohort.summarize_dataframe`)
        - provenance_file_summary: summary of provenance file contents (see `?cohorts.Cohort.summarize_provenance`)
        """
        provenance_file_summary = self.summarize_provenance()
        dataframe_hash = self.summarize_dataframe()
        results = {
            "provenance_file_summary": provenance_file_summary,
            "dataframe_hash": dataframe_hash
        }
        return(results)



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .variant_stats import variant_stats_from_variant

import numpy as np
from varcode import Substitution, Variant
from varcode.common import memoize
import re
import pandas as pd
from os import path

def no_filter(filterable_variant):
    return True

def variant_qc_filter(filterable_variant,
                      min_tumor_depth,
                      min_normal_depth,
                      min_tumor_vaf,
                      max_normal_vaf,
                      min_tumor_alt_depth):
    somatic_stats = variant_stats_from_variant(filterable_variant.variant,
                                               filterable_variant.variant_metadata)

    # Filter variant with depth < depth
    if (somatic_stats.tumor_stats.depth < min_tumor_depth or
        somatic_stats.normal_stats.depth < min_normal_depth):
        return False

    # Filter based on normal evidence
    if somatic_stats.normal_stats.variant_allele_frequency > max_normal_vaf:
        return False

    if somatic_stats.tumor_stats.variant_allele_frequency < min_tumor_vaf:
        return False

    if somatic_stats.tumor_stats.alt_depth < min_tumor_alt_depth:
        return False

    return True

@memoize
def expressed_variant_set(patient, variant_collection):
    # TODO: we're currently using the same isovar cache that we use for expressed
    # neoantigen prediction; so we pass in the same epitope lengths.
    # This is hacky and should be addressed.
    df_isovar = patient.cohort.load_single_patient_isovar(
        patient=patient,
        variants=variant_collection,
        epitope_lengths=[8, 9, 10, 11])
    expressed_variant_set = set()
    for _, row in df_isovar.iterrows():
        expressed_variant = Variant(contig=row["chr"],
                          start=row["pos"],
                          ref=row["ref"],
                          alt=row["alt"],
                          ensembl=variant_collection[0].ensembl)
        expressed_variant_set.add(expressed_variant)
    return expressed_variant_set

def variant_expressed_filter(filterable_variant, **kwargs):
    expressed_variants = expressed_variant_set(
        patient=filterable_variant.patient,
        variant_collection=filterable_variant.variant_collection)
    return filterable_variant.variant in expressed_variants

def effect_expressed_filter(filterable_effect, **kwargs):
    return variant_expressed_filter(filterable_effect, **kwargs)

def load_ensembl_coverage(cohort, coverage_path,
                          min_normal_depth, min_tumor_depth):
    """
    Load in Pageant CoverageDepth results with Ensembl loci.

    coverage_path is a path to Pageant CoverageDepth output directory, with
    one subdirectory per patient and a `cdf.csv` file inside each patient subdir.
    """
    columns = [
        "NormalDepth",
        "TumorDepth",
        "Normal BP",
        "Tumor BP",
        "Num Loci",
        "% Normal BP",
        "% Tumor BP",
        "% Loci",
        "Off-target Normal BP",
        "Off-target Tumor BP",
        "Off-target Num Loci",
        "Off-target % Normal BP",
        "Off-target % Tumor BP",
        "Off-target % Loci",
    ]
    ensembl_loci_dfs = []
    for patient in cohort:
        patient_ensembl_loci_df = pd.read_csv(
            path.join(coverage_path, patient.id, "cdf.csv"),
            names=columns)
        # pylint: disable=no-member
        # pylint gets confused by read_csv
        patient_ensembl_loci_df = patient_ensembl_loci_df[(
            (patient_ensembl_loci_df.NormalDepth == min_normal_depth) &
            (patient_ensembl_loci_df.TumorDepth == min_tumor_depth))]
        assert len(patient_ensembl_loci_df) == 1, (
            "Incorrect number of %d, %d depth loci results: %d" % (
                min_normal_depth, min_tumor_depth, len(patient_ensembl_loci_df)))
        patient_ensembl_loci_df["patient_id"] = patient.id
        ensembl_loci_dfs.append(patient_ensembl_loci_df)
    ensembl_loci_df = pd.concat(ensembl_loci_dfs)
    ensembl_loci_df["MB"] = ensembl_loci_df["Num Loci"] / 1000000.0
    return ensembl_loci_df[["patient_id", "Num Loci", "MB"]]



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import pandas as pd
from numpy.random import choice, randint, seed

from . import Cohort, Patient

def random_cohort(size, cache_dir, seed_val=1234):
    seed(seed_val)
    d = {}
    d["id"] = [str(id) for id in range(size)]
    d["age"] = choice([10, 15, 28, 32, 59, 62, 64, 66, 68], size)
    d["OS"] = [os + randint(10) for os in choice([10, 100, 500, 1000], size)]
    # Note: these values are not currently consistent with each other.
    d["PFS"] = [int(os * 0.6) for os in d["OS"]]
    d["benefit"] = choice([False, True], size)
    d["random"] = [randint(100) for i in range(size)]
    d["random_boolean"] = choice([False, True], size)
    d["benefit_correlate"] = [randint(50) if benefit else randint(20) for benefit in d["benefit"]]
    d["benefit_correlate_boolean"] = [True if corr > 10 else False for corr in d["benefit_correlate"]]
    d["deceased"] = choice([False, True], size)
    d["progressed_or_deceased"] = choice([False, True], size)
    df = pd.DataFrame(d)
    patients = []
    for i, row in df.iterrows():
         patient = Patient(
             id=row["id"],
             os=row["OS"],
             pfs=row["PFS"],
             benefit=row["benefit"],
             deceased=row["deceased"],
             progressed_or_deceased=row["progressed_or_deceased"],
             additional_data=row)
         patients.append(patient)
    return Cohort(
        patients=patients,
        cache_dir=cache_dir)



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from . import data_path, generated_data_path
from .data_generate import generate_vcfs
from .functions import *

from cohorts.cohort import filter_not_null

import pandas as pd
import numpy as np
from nose.tools import raises, eq_, ok_
from os import path, makedirs
from shutil import rmtree

from .test_basic import make_simple_cohort

def make_missing_vcf_cohort(patient_ids_with_missing_paths, missing_paths):
    file_format = "patient_format_%s.vcf"
    cohort = make_simple_cohort()
    patient_ids = [patient.id for patient in cohort]
    vcf_dir = generate_vcfs(id_to_mutation_count=dict(zip(patient_ids, [3, 3, 6])),
                            file_format=file_format,
                            template_name="vcf_template_1.vcf")
    for patient in cohort:
        vcf_paths = []
        vcf_filename = (file_format % patient.id)
        vcf_path = path.join(vcf_dir, vcf_filename)
        vcf_paths.append(vcf_path)
        if patient.id in patient_ids_with_missing_paths:
            patient.snv_vcf_paths = missing_paths
        else:
            patient.snv_vcf_paths = vcf_paths
        patient.hla_alleles = ["HLA-A02:01"]
    return vcf_dir, cohort

def test_broken_vcf_path():
    """
    Generate VCFs per-sample, and confirm that the NaNs are returned when
    VCF paths are broken.
    """
    vcf_dir, cohort = None, None
    try:
        vcf_dir, cohort = make_missing_vcf_cohort(
            patient_ids_with_missing_paths=["1"],
            missing_paths=["nonexistant_path"])

        count_col, df = cohort.as_dataframe(snv_count)
        eq_(len(df), 3)
        ok_(np.isnan(list(df[count_col])[0]))

        count_col, df = cohort.as_dataframe(missense_snv_count)
        eq_(len(df), 3)
        ok_(np.isnan(list(df[count_col])[0]))
    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()

def test_missing_vcf_path():
    """
    Generate VCFs per-sample, and confirm that the NaNs are returned when
    VCF paths are broken.
    """
    vcf_dir, cohort = None, None
    try:
        vcf_dir, cohort = make_missing_vcf_cohort(
            patient_ids_with_missing_paths=["1"],
            missing_paths=[])

        count_col, df = cohort.as_dataframe(snv_count)
        eq_(len(df), 3)
        ok_(np.isnan(list(df[count_col])[0]))

        count_col, df = cohort.as_dataframe(missense_snv_count)
        eq_(len(df), 3)
        ok_(np.isnan(list(df[count_col])[0]))
    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()

def test_no_neoantigens():
    """
    Make sure that exactly 0 neoantigens is treated differently than no calculated
    neoantigens.
    """
    vcf_dir, cohort = None, None
    try:
        vcf_dir, cohort = make_missing_vcf_cohort(
            patient_ids_with_missing_paths=["1"],
            missing_paths=[])

        generate_empty_neoantigens(cohort,
                                   patient_ids_with_zero_neoantigens=["4", "5"],
                                   patient_ids_with_empty_neoantigens=["1"])

        count_col, df = cohort.as_dataframe(neoantigen_count)
        eq_(len(df), 3)

        # This is used internally by plotting functions.
        # Ensure it filters NaN but not 0.
        eq_(len(filter_not_null(df, count_col)), 2)
    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()

def generate_empty_neoantigens(cohort,
                               patient_ids_with_zero_neoantigens,
                               patient_ids_with_empty_neoantigens):
    for patient in cohort:
        if patient.id in patient_ids_with_empty_neoantigens:
            continue
        elif patient.id in patient_ids_with_zero_neoantigens:
            neoantigen_path = generated_data_path(
                path.join("cache", "cached-neoantigens",
                          patient.id, "snv-union-neoantigens.csv"))
            if not path.exists(path.dirname(neoantigen_path)):
                makedirs(path.dirname(neoantigen_path))
                with open(neoantigen_path, "w") as f:
                    df_neoantigens = pd.read_csv(data_path("empty-neoantigens.csv"))
                    # pylint: disable=no-member
                    # pylint gets confused by to_csv
                    df_neoantigens.to_csv(neoantigen_path)
        else:
            raise ValueError("Patient ID needs to be empty or zero")



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from .utils import require_id_str

class Patient(object):
    """
    Represents a patient, which contains zero or more of: normal `Sample`,
    tumor `Sample`.

    Parameters
    __________
    id : str
        ID of the patient.
    os : int
        Overall survival in days.
    pfs : int
        Progression-free survival in days.
    deceased : bool
        Is the patient deceased?
    progressed : bool
        Has the patient progressed?
    progressed_or_deceased : bool
        Has the patient either progressed or passed away?
    benefit : bool
        Has the patient seen a durable clinical benefit?
    snv_vcf_paths : list
        List of paths to SNV VCFs or this patient; multple VCFs get merged.
    indel_vcf_paths : list
        List of paths to indel VCFs for this patient; multple VCFs get merged.
    normal_sample : Sample
        This patient's normal `Sample`.
    tumor_sample: Sample
        This patient's tumor `Sample`.
    hla_alleles : list
        A list of this patient's HLA class I alleles.
    additional_data : dict
        A dictionary of additional data: name of datum mapping to value.
    """
    def __init__(self,
                 id,
                 os,
                 pfs,
                 deceased,
                 progressed=None,
                 progressed_or_deceased=None,
                 benefit=None,
                 snv_vcf_paths=[],
                 indel_vcf_paths=[],
                 normal_sample=None,
                 tumor_sample=None,
                 hla_alleles=None,
                 additional_data=None,
                 cohort=None):
        require_id_str(id)
        self.id = id
        self.os = os
        self.pfs = pfs
        self.deceased = deceased
        self.progressed = progressed
        self.progressed_or_deceased = progressed_or_deceased
        self.benefit = benefit
        self.snv_vcf_paths = snv_vcf_paths
        self.indel_vcf_paths = indel_vcf_paths
        self.normal_sample = normal_sample
        self.tumor_sample = tumor_sample
        self.hla_alleles = hla_alleles
        self.additional_data = additional_data

        # TODO: This can be removed once all patient-specific functions are
        # removed from Cohort.
        self.cohort = cohort

        self.add_progressed_or_deceased()

    def add_progressed_or_deceased(self):
        assert self.progressed is not None or self.progressed_or_deceased is not None, (
            "Need at least one of progressed and progressed_or_deceased")
        if self.progressed_or_deceased is None:
            self.progressed_or_deceased = self.progressed or self.deceased

        # If we have both of these, ensure that they're in sync
        if self.progressed_or_deceased is not None and self.progressed is not None:
            assert self.progressed_or_deceased == self.progressed or self.deceased, (
                    "progressed_or_deceased should equal progressed || deceased")

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        if self is other:
            return True
        return (
            self.__class__ == other.__class__ and
            self.id == other.id)



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
from cohorts.utils import strip_column_names, _strip_column_name
from cohorts import DataFrameLoader, Cohort, Patient
import warnings

from . import generated_data_path
from .functions import *

from nose.tools import eq_, ok_
from .test_basic import make_simple_cohort

def make_alt_simple_clinical_dataframe(
        os_list=None,
        pfs_list=None,
        deceased_list=None,
        progressed_or_deceased_list=None):
    return pd.DataFrame({"id": ["1", "4", "5"],
                         "age": [15, 20, 25],
                         "os": [100, 150, 120] if os_list is None else os_list,
                         "pfs": [50, 40, 120] if pfs_list is None else pfs_list,
                         "deceased": [True, False, False] if deceased_list is None else deceased_list,
                         "progressed_or_deceased": [True, True, False] if progressed_or_deceased_list is None else progressed_or_deceased_list})

def make_alt_simple_cohort(merge_type="union", **kwargs):
    clinical_dataframe = make_alt_simple_clinical_dataframe(**kwargs)
    patients = []
    for i, row in clinical_dataframe.iterrows():
        patient = Patient(id=row["id"],
                          os=row["os"],
                          pfs=row["pfs"],
                          deceased=row["deceased"],
                          progressed_or_deceased=row["progressed_or_deceased"],
                          additional_data=row
                          )
        patients.append(patient)

    return Cohort(
        patients=patients,
        responder_pfs_equals_os=True,
        merge_type=merge_type,
        cache_dir=generated_data_path("cache"))


def test_strip_single_column_name():
    res = [_strip_column_name(col) for col in ['PD-L1', 'PD L1', 'PD L1_']]
    eq_(res, ['pd_l1', 'pd_l1', 'pd_l1'])


def test_strip_column_names():
    d = {'one': pd.Series([1., 2., 3.], index=['a', 'b', 'c']),
         'two': pd.Series([1., 2., 3., 4.], index=['a', 'b', 'c', 'd']),
         'PD L1 (val)': pd.Series([1., 2., 3., 4.], index=['a', 'b', 'c', 'd']),
         'PD L1 (>1)': pd.Series([0., 1., 1., 1.], index=['a', 'b', 'c', 'd']),
         }
    df = pd.DataFrame(d)

    # should not error & should rename columns
    df2 = df.rename(columns=strip_column_names(df.columns))
    ok_((df2.columns != df.columns).any())
    # should not rename columns -- should raise a warning
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        df3 = df.rename(columns=strip_column_names(
                        df.columns, keep_paren_contents=False))
        ok_(len(w) > 0, 'warning not raised when keep_paren_contents results in dups')
    ok_((df3.columns == df.columns).all())


def prep_test_cohort():
    cohort = make_simple_cohort()
    df_hello = pd.DataFrame({
        'one': [1., 2., 3.],
        'two': [1., 2., 3.],
        'PD L1 (val)': [1., 2., 3.],
        'PD L1 (>1)': [0., 1., 1.],
        'the_id': ['1', '5', '7'],
        })

    def load_df():
        return df_hello
    df_loader = DataFrameLoader("hello", load_df, join_on="the_id")
    cohort.df_loaders = [df_loader]
    return df_hello, cohort


def prep_alt_test_cohort():
    cohort = make_alt_simple_cohort()
    df_hello = pd.DataFrame({
        'one': [1., 2., 3.],
        'two': [1., 2., 3.],
        'PD L1 (val)': [1., 2., 3.],
        'PD L1 (>1)': [0., 1., 1.],
        'the_id': ['1', '5', '7'],
        })

    def load_df():
        return df_hello
    df_loader = DataFrameLoader("hello", load_df, join_on="the_id")
    cohort.df_loaders = [df_loader]
    return df_hello, cohort


def compare_column_names(expected, observed):
    expected = set(expected)
    observed = set(observed)
    print('Expected:', expected)
    print('Observed:', observed)
    return(expected.issubset(observed))


def test_as_dataframe_generic():
    df_hello, cohort = prep_test_cohort()
    # test that column names haven't changed
    df = cohort.as_dataframe(join_with="hello")
    # column names should match those in df_hello
    res = compare_column_names(expected = df_hello.columns,
                               observed = df.columns)
    ok_(res, 'columns names failed to match expected')


def test_as_dataframe_good_rename():
    df_hello, cohort = prep_alt_test_cohort()
    # test behavior with rename_cols=True. should not raise a warning
    df = cohort.as_dataframe(rename_cols=True, join_with='hello')
    res = compare_column_names(expected = strip_column_names(df_hello.columns),
                               observed = df.columns)
    ok_(res, 'column names failed to match expected')


def test_as_dataframe_bad_rename():
    df_hello, cohort = prep_test_cohort()
    # test behavior with rename_cols=True. should raise a warning
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        df = cohort.as_dataframe(rename_cols=True, join_with='hello')
        # skip test since warnings (for some reason) don't propagate
        #ok_(len(w) > 0, 'fail to generate dups warning when using rename_cols=True')
    res = compare_column_names(expected = df_hello.columns,
                               observed = df.columns)
    ok_(res, 'columns names failed to match expected')

def test_as_dataframe_drop_parens():
    df_hello, cohort = prep_test_cohort()
    # test behavior with keep_paren_contents=False
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        df = cohort.as_dataframe(rename_cols=True, keep_paren_contents=False, join_with='hello')
        # skip test for warning since warning doesn't propagate (not sure why)
        #ok_(len(w) > 0, 'no warning when duplicates resulting from rename_cols')
    res = compare_column_names(expected = df_hello.columns,
                               observed = df.columns)
    ok_(res, 'columns names failed to match expected')



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal, ROUND_HALF_UP

def round_float(f, digits, rounding=ROUND_HALF_UP):
    """
    Accurate float rounding from http://stackoverflow.com/a/15398691.
    """
    return Decimal(str(f)).quantize(Decimal(10) ** (-1 * digits),
                                    rounding=rounding)

def float_str(f, min_digits=2, max_digits=6):
    """
    Returns a string representing a float, where the number of
    significant digits is min_digits unless it takes more digits
    to hit a non-zero digit (and the number is 0 < x < 1).
    We stop looking for a non-zero digit after max_digits.
    """
    if f >= 1 or f <= 0:
        return str(round_float(f, min_digits))
    start_str = str(round_float(f, max_digits))
    digits = start_str.split(".")[1]
    non_zero_indices = []
    for i, digit in enumerate(digits):
        if digit != "0":
            non_zero_indices.append(i + 1)
    # Only saw 0s.
    if len(non_zero_indices) == 0:
        num_digits = min_digits
    else:
        # Of the non-zero digits, pick the num_digit'th of those (including any zeros)
        min_non_zero_indices = range(non_zero_indices[0], non_zero_indices[-1] + 1)[:min_digits]
        num_digits = min_non_zero_indices[-1]
    return str(round_float(f, num_digits))



from varcode import EffectCollection, VariantCollection, Variant

def genome(variant_collection):
    return variant_collection[0].ensembl

class FilterableVariant(object):
    def __init__(self, variant, variant_collection, patient):
        self.variant = variant
        self.variant_collection = variant_collection
        self.patient = patient

    @property
    def variant_metadata(self):
        return self.variant_collection.metadata[self.variant]

    @property
    def genome(self):
        return genome(self.variant_collection)

class FilterableEffect(FilterableVariant):
    def __init__(self, effect, variant_collection, patient):
        self.effect = effect
        FilterableVariant.__init__(self,
                                   variant=effect.variant,
                                   variant_collection=variant_collection,
                                   patient=patient)

class FilterableNeoantigen(FilterableVariant):
    def __init__(self, neoantigen_row, variant_collection, patient):
        self.neoantigen_row = neoantigen_row
        def build_variant(row, genome):
            return Variant(
                contig=row["chr"],
                ref=row["ref"],
                alt=row["alt"],
                start=row["start"],
                ensembl=genome)
        variant = build_variant(neoantigen_row, genome(variant_collection))
        FilterableVariant.__init__(self,
                                   variant=variant,
                                   variant_collection=variant_collection,
                                   patient=patient)

class FilterablePolyphen(FilterableVariant):
    def __init__(self, polyphen_row, variant_collection, patient):
        self.polyphen_row = polyphen_row
        def build_variant(row, genome):
            return Variant(
                contig=row["chrom"],
                ref=row["ref"],
                alt=row["alt"],
                start=row["pos"],
                ensembl=genome)
        variant = build_variant(polyphen_row, genome(variant_collection))
        FilterableVariant.__init__(self,
                                   variant=variant,
                                   variant_collection=variant_collection,
                                   patient=patient)

def filter_variants(variant_collection, patient, filter_fn):
    """Filter variants from the Variant Collection

    Parameters
    ----------
    variant_collection : varcode.VariantCollection
    patient : cohorts.Patient
    filter_fn: function
        Takes a FilterableVariant and returns a boolean. Only variants returning True are preserved.

    Returns
    -------
    varcode.VariantCollection
        Filtered variant collection, with only the variants passing the filter
    """
    if filter_fn:
        return variant_collection.clone_with_new_elements([
            variant
            for variant in variant_collection
            if filter_fn(FilterableVariant(
                    variant=variant,
                    variant_collection=variant_collection,
                    patient=patient))
        ])
    else:
        return variant_collection

def filter_effects(effect_collection, variant_collection, patient, filter_fn):
    """Filter variants from the Effect Collection

    Parameters
    ----------
    effect_collection : varcode.EffectCollection
    variant_collection : varcode.VariantCollection
    patient : cohorts.Patient
    filter_fn: function
        Takes a FilterableEffect and returns a boolean. Only effects returning True are preserved.

    Returns
    -------
    varcode.EffectCollection
        Filtered effect collection, with only the variants passing the filter
    """
    if filter_fn:
        return EffectCollection([
            effect
            for effect in effect_collection
            if filter_fn(FilterableEffect(
                    effect=effect,
                    variant_collection=variant_collection,
                    patient=patient))])
    else:
        return effect_collection

def filter_neoantigens(neoantigens_df, variant_collection, patient, filter_fn):
    if filter_fn:
        filter_mask = neoantigens_df.apply(
            lambda row: filter_fn(
                FilterableNeoantigen(neoantigen_row=row,
                                     variant_collection=variant_collection,
                                     patient=patient)), axis=1)
        return neoantigens_df[filter_mask]
    else:
        return neoantigens_df

def filter_polyphen(polyphen_df, variant_collection, patient, filter_fn):
    if filter_fn:
        filter_mask = polyphen_df.apply(
            lambda row: filter_fn(
                FilterablePolyphen(polyphen_row=row,
                                   variant_collection=variant_collection,
                                   patient=patient)), axis=1)
        return polyphen_df[filter_mask]
    else:
        return polyphen_df



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import pandas as pd
from nose.tools import eq_, ok_

from cohorts.plot import mann_whitney_plot

def test_mann_whitney():
    data = pd.DataFrame({"distribution": range(1, 7),
                         "condition": [True, False] * 3})
    U_default, p_default, _, _, _, _ = mann_whitney_plot(
        data, "condition", "distribution", skip_plot=True)
    U_two, p_two, _, _, _, _ = mann_whitney_plot(
        data, "condition", "distribution", alternative="two-sided", skip_plot=True)
    U_less, p_less, _, _, _, _ = mann_whitney_plot(
        data, "condition", "distribution", alternative="less", skip_plot=True)
    U_greater, p_greater, _, _, _, _ = mann_whitney_plot(
        data, "condition", "distribution", alternative="greater", skip_plot=True)
    eq_(p_default, p_two)
    ok_((p_default == p_less * 2) or (p_default == p_greater * 2))



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from nose.tools import eq_, ok_
import pandas as pd
from os import path
import os
from shutil import rmtree
import cohorts
from cohorts.provenance import compare_provenance
import warnings

from .test_basic import make_simple_cohort

def test_provenance():
    cohort = None
    try:
        # First, let's just make a cohort and cache a file. Before we do this,
        # we have no provenance.
        cohort = make_simple_cohort()
        cohort.check_provenance = True
        df_empty = pd.DataFrame({"a": [1]})
        patient_id = "1"
        cache_name = cohort.cache_names["variant"]
        cache_dir = path.join(cohort.cache_dir, cache_name)
        patient_cache_dir = path.join(cache_dir, str(patient_id))
        provenance_path = path.join(patient_cache_dir, "PROVENANCE")
        ok_(not path.exists(provenance_path))

        # After we cache the file, we should have provenance.
        cohort.save_to_cache(df_empty, cache_name, patient_id, "cached_file.csv")
        ok_(path.exists(provenance_path))

        # Now let's mess with the cached provenance file.
        df_empty = cohort.load_from_cache(cache_name, patient_id, "cached_file.csv")
        provenance = cohort.load_provenance(patient_cache_dir)
        provenance["hello"] = "1.0.1"
        cohort.save_provenance(patient_cache_dir, provenance)

        # We should get a warning because our saved provenance is different
        # from our current environment.
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            cohort.load_from_cache(cache_name, patient_id, "cached_file.csv")
            eq_(len(w), 1)

        # But, when check_provenance is off, we shouldn't get a warning.
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            cohort.check_provenance = False
            cohort.load_from_cache(cache_name, patient_id, "cached_file.csv")
            eq_(len(w), 0)
    finally:
        if cohort is not None:
            cohort.clear_caches()

def test_summarize_provenance():
    cohort = None
    try:
        # First, let's just make a cohort and cache a file. Before we do this,
        # we have no provenance.
        cohort = make_simple_cohort()
        cohort.check_provenance = True
        df_empty = pd.DataFrame({"a": [1]})
        patient_id = "1"
        cache_name = cohort.cache_names["variant"]
        cache_dir = path.join(cohort.cache_dir, cache_name)
        patient_cache_dir = path.join(cache_dir, str(patient_id))
        provenance_path = path.join(patient_cache_dir, "PROVENANCE")
        ok_(not path.exists(provenance_path))
        ## should not have any provenance summary (but shouldn't fail) if no provenance 
        summary = cohort.summarize_provenance()

        # After we cache the file, we should have provenance.
        cohort.save_to_cache(df_empty, cache_name, patient_id, "cached_file.csv")
        ## and, we should be able to summarize provenance
        summary = cohort.summarize_provenance()

        ## add second patient id
        patient_id = "4" ## is within patient ids generated by make_simple_cohort
        patient_cache_dir = path.join(cache_dir, str(patient_id))
        provenance_path = path.join(patient_cache_dir, "PROVENANCE")
        ok_(not path.exists(provenance_path))
        ## shouldn't fail (some patients have provenance & some don't)
        summary = cohort.summarize_provenance()
        ## save to file 
        cohort.save_to_cache(df_empty, cache_name, patient_id, "cached_file.csv")
        ## shouldn't fail (both patients have provenance & should be equivalent)
        summary = cohort.summarize_provenance()

        # Now let's mess with the second provenance file.
        df_empty = cohort.load_from_cache(cache_name, patient_id, "cached_file.csv")
        provenance = cohort.load_provenance(patient_cache_dir)
        provenance["pandas"] = "1.0.1"
        print(provenance)
        print("\n")
        cohort.save_provenance(patient_cache_dir, provenance)
        ## confirm provenance was updated
        new_provenance = cohort.load_provenance(patient_cache_dir)
        ok_(compare_provenance(provenance, new_provenance) == 0)
        print(new_provenance)

        ## should see a warning when loading provenance
        ## & provenance value should be None
        # But, when check_provenance is off, we shouldn't get a warning.
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            ok_(not cohort.summarize_provenance_per_cache()[cache_name])
            ok_(len(w)>0)

        ## now test whether deleting contents of cache_dir causes an error
        dirlist = [f for f in os.listdir(cache_dir)]
        for dir in dirlist:
            rmtree(path.join(cache_dir, dir))
        summary = cohort.summarize_provenance()
    finally:
        if cohort is not None:
            cohort.clear_caches()



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import matplotlib as mpl
import matplotlib.colors as colors
import seaborn as sb
import numpy as np

def set_styling():
    sb.set_style("white")
    red = colors.hex2color("#bb3f3f")
    blue = colors.hex2color("#5a86ad")
    deep_colors = sb.color_palette("deep")
    green = deep_colors[1]
    custom_palette = [red, blue, green]
    custom_palette.extend(deep_colors[3:])
    sb.set_palette(custom_palette)
    mpl.rcParams.update({"figure.figsize": np.array([6, 6]),
                         "legend.fontsize": 12,
                         "font.size": 16,
                         "axes.labelsize": 16,
                         "axes.labelweight": "bold",
                         "xtick.labelsize": 16,
                         "ytick.labelsize": 16})



from collections import namedtuple


VariantStats = namedtuple("VariantStats",
                          ["depth", "alt_depth", "variant_allele_frequency"])


SomaticVariantStats = namedtuple("SomaticVariantStats",
                          ["tumor_stats", "normal_stats"])


def strelka_somatic_variant_stats(variant, variant_metadata):
    """Parse out the variant calling statistics for a given variant from a Strelka VCF

    Parameters
    ----------
    variant : varcode.Variant
    sample_info : dict
        Dictionary of sample to variant calling statistics, corresponds to the sample columns
        in a Strelka VCF

    Returns
    -------
    SomaticVariantStats
    """

    sample_info = variant_metadata["sample_info"]
    # Ensure there are exactly two samples in the VCF, a tumor and normal
    assert len(sample_info) == 2, "More than two samples found in the somatic VCF"

    tumor_stats = _strelka_variant_stats(variant, sample_info["TUMOR"])
    normal_stats = _strelka_variant_stats(variant, sample_info["NORMAL"])
    return SomaticVariantStats(tumor_stats=tumor_stats, normal_stats=normal_stats)

def _strelka_variant_stats(variant, sample_info):
    """Parse a single sample"s variant calling statistics based on Strelka VCF output

    Parameters
    ----------
    variant : varcode.Variant
    sample_info : dict
        Dictionary of Strelka-specific variant calling fields

    Returns
    -------
    VariantStats
    """

    # Retrieve the Tier 1 counts from Strelka
    ref_depth = int(sample_info[variant.ref+"U"][0])
    alt_depth = int(sample_info[variant.alt+"U"][0])
    depth = alt_depth + ref_depth
    vaf = float(alt_depth) / depth

    return VariantStats(depth=depth, alt_depth=alt_depth, variant_allele_frequency=vaf)

def mutect_somatic_variant_stats(variant, variant_metadata):
    """Parse out the variant calling statistics for a given variant from a Mutect VCF

    Parameters
    ----------
    variant : varcode.Variant
    sample_info : dict
        Dictionary of sample to variant calling statistics, corresponds to the sample columns
        in a Mutect VCF

    Returns
    -------
    SomaticVariantStats
    """

    sample_info = variant_metadata["sample_info"]
    # Ensure there are exactly two samples in the VCF, a tumor and normal
    assert len(sample_info) == 2, "More than two samples found in the somatic VCF"

    # Find the sample with the genotype field set to variant in the VCF
    tumor_sample_infos = [info for info in sample_info.values() if info["GT"] == "0/1"]

    # Ensure there is only one such sample
    assert len(tumor_sample_infos) == 1, "More than one tumor sample found in the VCF file"

    tumor_sample_info = tumor_sample_infos[0]
    normal_sample_info = [info for info in sample_info.values() if info["GT"] != "0/1"][0]

    tumor_stats = _mutect_variant_stats(variant, tumor_sample_info)
    normal_stats = _mutect_variant_stats(variant, normal_sample_info)
    return SomaticVariantStats(tumor_stats=tumor_stats, normal_stats=normal_stats)

def _mutect_variant_stats(variant, sample_info):
    """Parse a single sample"s variant calling statistics based on Mutect"s (v1) VCF output

    Parameters
    ----------
    variant : varcode.Variant
    sample_info : dict
        Dictionary of Mutect-specific variant calling fields

    Returns
    -------
    VariantStats
    """

    # Parse out the AD (or allele depth field), which is an array of [REF_DEPTH, ALT_DEPTH]
    ref_depth, alt_depth = sample_info["AD"]
    depth = int(ref_depth) + int(alt_depth)
    vaf = float(alt_depth) / depth

    return VariantStats(depth=depth, alt_depth=alt_depth, variant_allele_frequency=vaf)


def variant_stats_from_variant(variant,
                               metadata,
                               merge_fn=(lambda all_stats: \
                                max(all_stats, key=(lambda stats: stats.tumor_stats.depth)))):
    """Parse the variant calling stats from a variant called from multiple VCFs. The stats are merged
    based on `merge_fn`

    Parameters
    ----------
    variant : varcode.Variant
    metadata : dict
        Dictionary of VCF to variant calling metadata from that VCF file
    merge_fn : function
        Function from list of SomaticVariantStats to single SomaticVariantStats.
        This is used if a variant is called by multiple callers or appears in multiple VCFs.
        By default, this uses the data from the caller that had a higher tumor depth.

    Returns
    -------
    SomaticVariantStats
    """

    all_stats = []
    for (vcf, variant_metadata) in metadata.items():
        if "strelka" in vcf.lower():
            stats = strelka_somatic_variant_stats(variant, variant_metadata)
        elif "mutect" in vcf.lower():
            stats = mutect_somatic_variant_stats(variant, variant_metadata)
        else:
            raise ValueError("Cannot parse sample fields, VCF file {} is from an unsupported caller.".format(vcf))
        all_stats.append(stats)

    return merge_fn(all_stats)



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from varcode import ExonicSpliceSite, Substitution
from cohorts.variant_filters import no_filter

from .data_generate import generate_vcfs
from .functions import *

from nose.tools import raises, eq_, ok_
from mock import MagicMock
from os import path
from shutil import rmtree
from collections import defaultdict
import pandas as pd

from .test_basic import make_simple_cohort

FILE_FORMAT_1 = "patient_format1_%s.vcf"
FILE_FORMAT_2 = "patient_format2_%s.vcf"
FILE_FORMAT_3 = "patient_format3_%s.vcf"

def make_cohort(file_formats, merge_type="union"):
    cohort = make_simple_cohort(merge_type=merge_type)
    patient_ids = [patient.id for patient in cohort]
    vcf_dir = generate_vcfs(id_to_mutation_count=dict(zip(patient_ids, [3, 3, 6])),
                            file_format=FILE_FORMAT_1,
                            template_name="vcf_template_1.vcf")
    _ = generate_vcfs(id_to_mutation_count=dict(zip(patient_ids, [4, 1, 5])),
                      file_format=FILE_FORMAT_2,
                      template_name="vcf_template_1.vcf")
    _ = generate_vcfs(id_to_mutation_count=dict(zip(patient_ids, [5, 2, 3])),
                      file_format=FILE_FORMAT_3,
                      template_name="vcf_template_2.vcf")
    for patient in cohort:
        vcf_paths = []
        for file_format in file_formats:
            vcf_filename = (file_format % patient.id)
            vcf_path = path.join(vcf_dir, vcf_filename)
            vcf_paths.append(vcf_path)
        patient.snv_vcf_paths = vcf_paths
    return vcf_dir, cohort

def test_snv_counts():
    """
    Generate VCFs per-sample, and confirm that the counting functions work as expected.
    """
    vcf_dir, cohort = None, None
    try:
        # Use all three VCF sources
        vcf_dir, cohort = make_cohort([FILE_FORMAT_1])

        # The SNV count should be exactly what we generated
        count_col, df = cohort.as_dataframe(snv_count)
        eq_(len(df), 3)
        eq_(list(df[count_col]), [3, 3, 6])

        count_col, df = cohort.as_dataframe(missense_snv_count)
        eq_(len(df), 3)
        eq_(list(df[count_col]), [2, 2, 4])
    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()

def test_merge_three():
    """
    Generate three VCFs per-sample and confirm that merging works as expected.
    """
    vcf_dir, cohort = None, None
    try:
        # Use all three VCF sources
        vcf_dir, cohort = make_cohort([FILE_FORMAT_1, FILE_FORMAT_2, FILE_FORMAT_3],
                                      merge_type="union")

        # [3, 3, 6] and [4, 1, 5] use the same template, resulting in a union of [4, 3, 6] unique variants
        # [5, 2, 3] uses a separate template, resulting in a union of [4, 3, 6] + [5, 2, 3] = [9, 5, 9] unique variants
        count_col, df = cohort.as_dataframe(snv_count)
        eq_(len(df), 3)
        eq_(list(df[count_col]), [9, 5, 9])

        # For intersection, variants need to appear in *all*, here. None of them do.
        vcf_dir, cohort = make_cohort([FILE_FORMAT_1, FILE_FORMAT_2, FILE_FORMAT_3],
                                      merge_type="intersection")
        count_col, df = cohort.as_dataframe(snv_count)
        eq_(list(df[count_col]), [0, 0, 0])
    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()

def test_merge_two():
    """
    Generate two VCFs per-sample and confirm that merging works as expected.
    """
    vcf_dir, cohort = None, None
    try:
        # Now, with only two VCF sources
        vcf_dir, cohort = make_cohort([FILE_FORMAT_1, FILE_FORMAT_2],
                                      merge_type="union")

        # [3, 3, 6] and [4, 1, 5] use the same template, resulting in a union of [4, 3, 6] unique variants
        count_col, df = cohort.as_dataframe(snv_count)
        eq_(len(df), 3)
        eq_(list(df[count_col]), [4, 3, 6])

        # For intersection, some variants do appear in both.
        vcf_dir, cohort = make_cohort([FILE_FORMAT_1, FILE_FORMAT_2],
                                      merge_type="intersection")
        count_col, df = cohort.as_dataframe(snv_count)
        eq_(len(df), 3)
        eq_(list(df[count_col]), [3, 1, 5])

        cohort_variants = cohort.load_variants(filter_fn=None)
        for (sample, variants) in cohort_variants.items():
            for variant in variants:
                metadata = variants.metadata[variant]
                eq_(len(metadata), 2) # Each variant has two metadata entries

    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()

def test_filter_variants():
    vcf_dir, cohort = None, None
    try:
        vcf_dir, cohort = make_cohort([FILE_FORMAT_1, FILE_FORMAT_2])

        def filter_g_variants(filterable_variant):
            return filterable_variant.variant.ref == 'G'
        g_variants = {'1': 2, '4': 1, '5': 3}

        cohort_variants = cohort.load_variants(filter_fn=filter_g_variants)

        for (sample, variants) in cohort_variants.items():
            eq_(len(variants), g_variants[sample])

    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()

def test_filter_effects():
    vcf_dir, cohort = None, None
    try:
        vcf_dir, cohort = make_cohort([FILE_FORMAT_1])

        def filter_substitution_effects(filterable_effect):
            return type(filterable_effect.effect) == Substitution
        missense_counts = {'1': 2, '4': 2, '5': 4}

        cohort_effects = cohort.load_effects(only_nonsynonymous=True, filter_fn=filter_substitution_effects)
        for (sample, effects) in cohort_effects.items():
            eq_(len(effects), missense_counts[sample])

        def filter_exonic_splice_site_effects(filterable_effect):
            return type(filterable_effect.effect) == ExonicSpliceSite
        splice_site_counts = {'1': 1, '4': 1, '5': 2}

        cohort_effects = cohort.load_effects(only_nonsynonymous=True, filter_fn=filter_exonic_splice_site_effects)
        for (sample, effects) in cohort_effects.items():
            eq_(len(effects), splice_site_counts[sample])

    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()

def test_multiple_effects():
    vcf_dir, cohort = None, None
    try:
        vcf_dir, cohort = make_cohort([FILE_FORMAT_1])
        effects = cohort.load_effects(only_nonsynonymous=False)
        for patient in cohort:
            variant_to_effect_set = defaultdict(set)
            for effect in effects[patient.id]:
                variant_to_effect_set[effect.variant].add(effect)
            for variant, effect_set in variant_to_effect_set.items():
                eq_(len(effect_set), 1,
                    "Variant %s should only have 1 effect but it has %s"
                    % (variant, len(effect_set)))
    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()

def test_cohort_default():
    """
    Checks that the Cohort default is used when intended, and not used when overridden.
    """
    vcf_dir, cohort = None, None
    try:
        vcf_dir, cohort = make_cohort([FILE_FORMAT_1])

        def default_filter_fn(filterable_variant):
            return filterable_variant.variant.ref == "A"

        # Mock out load_single_patient_isovar since we don't actually have that data,
        # yet it's needed for expressed_missense_snv_count.
        variants = cohort.load_variants()
        isovar_mocked = {}
        for patient in cohort:
            chr_list = []
            pos_list = []
            ref_list = []
            alt_list = []
            patient_variants = variants[patient.id]
            for variant in patient_variants:
                chr_list.append(variant.contig)
                pos_list.append(variant.start)
                ref_list.append(variant.ref)
                alt_list.append(variant.alt)
            df_isovar_mocked = pd.DataFrame({"chr": chr_list,
                                             "pos": pos_list,
                                             "ref": ref_list,
                                             "alt": alt_list})
            isovar_mocked[patient.id] = df_isovar_mocked

        def mock_function(patient, **kwargs):
            return isovar_mocked[patient.id]

        cohort.load_single_patient_isovar = MagicMock(side_effect=mock_function)

        # Especially need to include expressed_missense_snv_count, which works a little differently.
        for count_func in [snv_count, missense_snv_count, expressed_missense_snv_count]:
            # No Cohort default
            cohort.filter_fn = None
            count_col_no_arg, df_no_arg  = cohort.as_dataframe(count_func)
            count_col_default_arg, df_default_arg = cohort.as_dataframe(count_func, filter_fn=default_filter_fn)
            count_col_none_arg, df_none_arg = cohort.as_dataframe(count_func, filter_fn=None)
            count_col_no_filter_arg, df_no_filter_arg = cohort.as_dataframe(count_func, filter_fn=no_filter)
            ok_((df_no_arg[count_col_no_arg] == df_none_arg[count_col_none_arg]).all())
            ok_((df_default_arg[count_col_default_arg] != df_no_arg[count_col_no_arg]).any())
            ok_((df_no_filter_arg[count_col_no_filter_arg] == df_no_arg[count_col_no_arg]).all())

            # With a Cohort default
            cohort.filter_fn = default_filter_fn
            count_col_no_arg, df_no_arg  = cohort.as_dataframe(count_func)
            count_col_default_arg, df_default_arg = cohort.as_dataframe(count_func, filter_fn=default_filter_fn)
            count_col_none_arg, df_none_arg = cohort.as_dataframe(count_func, filter_fn=None)
            count_col_no_filter_arg, df_no_filter_arg = cohort.as_dataframe(count_func, filter_fn=no_filter)
            ok_((df_no_arg[count_col_no_arg] == df_none_arg[count_col_none_arg]).all())
            ok_((df_default_arg[count_col_default_arg] == df_no_arg[count_col_no_arg]).all())
            ok_((df_no_filter_arg[count_col_no_filter_arg] != df_no_arg[count_col_no_arg]).any())
    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

#pylint: disable=no-name-in-module
from nose.tools import assert_almost_equals, eq_, ok_, raises
from shutil import rmtree
from os import path

from .data_generate import generate_vcfs
from .test_basic import make_simple_cohort
from cohorts.variant_stats import (strelka_somatic_variant_stats, 
                                   mutect_somatic_variant_stats,
                                   variant_stats_from_variant)

FILE_FORMAT_1 = "patient_format1_%s.mutect.vcf"
FILE_FORMAT_2 = "patient_format2_%s.strelka.vcf"
FILE_FORMAT_3 = "patient_format3_%s.vcf"

FILE_FORMAT_4 = "patient_format4_%s.mutect.vcf"

def make_cohort(file_formats):
    cohort = make_simple_cohort()
    patient_ids = [patient.id for patient in cohort]

    # Mutect VCF format
    vcf_dir = generate_vcfs(id_to_mutation_count=dict(zip(patient_ids, [3, 3, 6])),
                            file_format=FILE_FORMAT_1,
                            template_name="vcf_template_1.vcf")

    # Strelka VCF format
    _ = generate_vcfs(id_to_mutation_count=dict(zip(patient_ids, [5, 2, 3])),
                      file_format=FILE_FORMAT_2,
                      template_name="vcf_template_2.vcf")

    # Strelka VCF format, but without Strelka in the name (mimic unsupported)
    _ = generate_vcfs(id_to_mutation_count=dict(zip(patient_ids, [5, 2, 3])),
                      file_format=FILE_FORMAT_3,
                      template_name="vcf_template_2.vcf")

    # Mutect VCF format with overlapping variants from the Strelka set (template 2)
    _ = generate_vcfs(id_to_mutation_count=dict(zip(patient_ids, [3, 2, 3])),
                      file_format=FILE_FORMAT_4,
                      template_name="vcf_template_3.vcf")

    for patient in cohort:
        vcf_paths = []
        for file_format in file_formats:
            vcf_filename = (file_format % patient.id)
            vcf_path = path.join(vcf_dir, vcf_filename)
            vcf_paths.append(vcf_path)
        patient.snv_vcf_paths = vcf_paths
    return vcf_dir, cohort

def extraction(cohort, extractor):
    variants = cohort.load_variants(filter_fn=None)
    for (sample, sample_variants) in variants.items():
        ok_(len(sample_variants) > 0)
        for variant in sample_variants:
            somatic_stats = extractor(variant, sample_variants.metadata[variant])

            ok_(somatic_stats.tumor_stats.depth > 10)
            ok_(somatic_stats.normal_stats.depth > 10)

            ok_(somatic_stats.tumor_stats.variant_allele_frequency > 0.0)
            ok_(somatic_stats.tumor_stats.variant_allele_frequency < 0.5)

            ok_(somatic_stats.normal_stats.variant_allele_frequency == 0)

@raises(ValueError)
def test_extract_unsupported_stats():
    vcf_dir, cohort = None, None
    try:
        # Use Strelka and unsupported VCF format
        vcf_dir, cohort = make_cohort([FILE_FORMAT_2, FILE_FORMAT_3])
        extraction(cohort, extractor=variant_stats_from_variant)

    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()

def test_extract_strelka_stats():
    vcf_dir, cohort = None, None
    try:
        # Use Strelka VCF format
        vcf_dir, cohort = make_cohort([FILE_FORMAT_2])
        extraction(cohort, extractor=strelka_somatic_variant_stats)

    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()

def test_extract_mutect_stats():
    vcf_dir, cohort = None, None
    try:
        # Use Mutect VCF format
        vcf_dir, cohort = make_cohort([FILE_FORMAT_1])
        extraction(cohort, extractor=mutect_somatic_variant_stats)

    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()

def test_extract_strelka_mutect_stats():
    vcf_dir, cohort = None, None
    try:
        # Use Mutect and Strelka VCF format
        vcf_dir, cohort = make_cohort([FILE_FORMAT_1, FILE_FORMAT_2])
        extraction(cohort, extractor=variant_stats_from_variant)

    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()

def test_extract_and_merge_strelka_mutect_stats():
    vcf_dir, cohort = None, None
    try:
        # Use Mutect and Strelka VCF format with overlapping variants
        vcf_dir, cohort = make_cohort([FILE_FORMAT_2, FILE_FORMAT_4])
        variants = cohort.load_variants(filter_fn=None)
        patient_1_variants = variants['1']

        def test_stats(variant_index, expected_tumor_depth, expected_tumor_vaf):
            variant = patient_1_variants[variant_index]
            stats = variant_stats_from_variant(variant, patient_1_variants.metadata[variant])
            eq_(stats.tumor_stats.depth, expected_tumor_depth)
            assert_almost_equals(stats.tumor_stats.variant_allele_frequency, expected_tumor_vaf, places=3)

        # The first 3 variants (all on chr12) are in both Mutect and Strelka VCFs
        test_stats(0, expected_tumor_depth=133, expected_tumor_vaf=0.038) # Higher depth in Mutect file
        test_stats(1, expected_tumor_depth=144, expected_tumor_vaf=0.0486) # Higher depth in Strelka file
        test_stats(2, expected_tumor_depth=45, expected_tumor_vaf=0.089) # Higher depth in Mutect file

        # The next 2 variants only appear in the Strelka VCFs
        test_stats(3, expected_tumor_depth=104, expected_tumor_vaf=0.058)
        test_stats(4, expected_tumor_depth=117, expected_tumor_vaf=0.077)

    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

class Sample(object):
    """
    Represents a single tumor or normal sample. It can point to DNA and/or
    RNA reads.

    Parameters
    __________
    is_tumor : bool
        Does this `Sample` represent a tumor sample?
    bam_path_dna : str
        Path to the DNA BAM file.
    bam_path_rna : str
        Path to the RNA BAM file.
    cufflinks_path : str
        Path to the Cufflinks output file.
    """
    def __init__(self,
                 is_tumor,
                 bam_path_dna=None,
                 bam_path_rna=None,
                 cufflinks_path=None,
                 kallisto_path=None):
        self.is_tumor = is_tumor
        self.bam_path_dna = bam_path_dna
        self.bam_path_rna = bam_path_rna
        self.cufflinks_path = cufflinks_path
        self.kallisto_path = kallisto_path



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from . import data_path, generated_data_path, DATA_DIR
from .data_generate import generate_vcfs
from .functions import *

from cohorts import Cohort, Patient
from cohorts.utils import InvalidDataError

import pandas as pd
from nose.tools import raises, eq_, ok_

def make_simple_clinical_dataframe(
        os_list=None,
        pfs_list=None,
        deceased_list=None,
        progressed_or_deceased_list=None):
    return pd.DataFrame({"id": ["1", "4", "5"],
                         "age": [15, 20, 25],
                         "OS": [100, 150, 120] if os_list is None else os_list,
                         "PFS": [50, 40, 120] if pfs_list is None else pfs_list,
                         "deceased": [True, False, False] if deceased_list is None else deceased_list,
                         "progressed_or_deceased": [True, True, False] if progressed_or_deceased_list is None else progressed_or_deceased_list})

def make_simple_cohort(merge_type="union", **kwargs):
    clinical_dataframe = make_simple_clinical_dataframe(**kwargs)
    patients = []
    for i, row in clinical_dataframe.iterrows():
        patient = Patient(id=row["id"],
                          os=row["OS"],
                          pfs=row["PFS"],
                          deceased=row["deceased"],
                          progressed_or_deceased=row["progressed_or_deceased"],
                          additional_data=row
                          )
        patients.append(patient)

    return Cohort(
        patients=patients,
        responder_pfs_equals_os=True,
        merge_type=merge_type,
        cache_dir=generated_data_path("cache"))

def test_pfs_equal_to_os():
    # Should not error
    make_simple_cohort(pfs_list=[100, 150, 120])

@raises(InvalidDataError)
def test_pfs_greater_than_os():
    make_simple_cohort(pfs_list=[120, 150, 120])

@raises(InvalidDataError)
def test_progressed_vs_pfs():
    make_simple_cohort(progressed_or_deceased_list=[True, False, False])

def test_simple_cohort():
    cohort = make_simple_cohort()
    eq_(len(cohort.as_dataframe()), 3)

    columns = set(cohort.as_dataframe().columns)
    ok_("id" in columns)
    ok_("patient_id" in columns)
    ok_("age" in columns)
    ok_("pfs" in columns)
    ok_("os" in columns)



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import re
import warnings

class InvalidDataError(ValueError):
    pass

def first_not_none_param(params, default):
    """
    Given a list of `params`, use the first param in the list that is
    not None. If all are None, fall back to `default`.
    """
    for param in params:
        if param is not None:
            return param
    return default

def filter_not_null(df, col):
    original_len = len(df)
    df = df[df[col].notnull()]
    updated_len = len(df)
    if updated_len < original_len:
        print("Missing %s for %d patients: from %d to %d" % (col, original_len - updated_len, original_len, updated_len))
    return df

def require_id_str(id):
    if type(id) != str:
        raise ValueError("Expected ID string, but id = %s" % str(id))

def _strip_column_name(col_name, keep_paren_contents=True):
    """
    Utility script applying several regexs to a string.
    Intended to be used by `strip_column_names`.

    This function will:
        1. replace informative punctuation components with text
        2. (optionally) remove text within parentheses
        3. replace remaining punctuation/whitespace with _
        4. strip leading/trailing punctuation/whitespace

    Parameters
    ----------
    col_name (str): input character string
    keep_paren_contents (logical):
        controls behavior of within-paren elements of text
         - if True, (the default) all text within parens retained
         - if False, text within parens will be removed from the field name

    Returns
    --------
    modified string for new field name

    Examples
    --------
    > print([_strip_column_name(col) for col in ['PD-L1','PD L1','PD L1_']])
    """
    # start with input
    new_col_name = col_name
    # replace meaningful punctuation with text equivalents
    # surround each with whitespace to enforce consistent use of _
    punctuation_to_text = {
        '<=': 'le',
        '>=': 'ge',
        '=<': 'le',
        '=>': 'ge',
        '<': 'lt',
        '>': 'gt',
        '#': 'num'
    }
    for punctuation, punctuation_text in punctuation_to_text.items():
        new_col_name = new_col_name.replace(punctuation, punctuation_text)

    # remove contents within ()
    if not(keep_paren_contents):
        new_col_name = re.sub('\([^)]*\)', '', new_col_name)

    # replace remaining punctuation/whitespace with _
    punct_pattern = '[\W_]+'
    punct_replacement = '_'
    new_col_name = re.sub(punct_pattern, punct_replacement, new_col_name)

    # remove leading/trailing _ if it exists (if last char was punctuation)
    new_col_name = new_col_name.strip("_")

    # TODO: check for empty string

    # return lower-case version of column name
    return new_col_name.lower()

def strip_column_names(cols, keep_paren_contents=True):
    """
    Utility script for renaming pandas columns to patsy-friendly names.

    Revised names have been:
        - stripped of all punctuation and whitespace (converted to text or `_`)
        - converted to lower case

    Takes a list of column names, returns a dict mapping
    names to revised names.

    If there are any concerns with the conversion, this will
    print a warning & return original column names.

    Parameters
    ----------

    cols (list): list of strings containing column names
    keep_paren_contents (logical):
        controls behavior of within-paren elements of text
         - if True, (the default) all text within parens retained
         - if False, text within parens will be removed from the field name

    Returns
    -------

    dict mapping col_names -> new_col_names

    Example
    -------

    > df = {'one' : pd.Series([1., 2., 3.], index=['a', 'b', 'c']),
      'two' : pd.Series([1., 2., 3., 4.], index=['a', 'b', 'c', 'd']),
      'PD L1 (value)': pd.Series([1., 2., 3., 4.], index=['a', 'b', 'c', 'd']),
      'PD L1 (>1)': pd.Series([0., 1., 1., 0.], index=['a', 'b', 'c', 'd']),
      }
    > df = pd.DataFrame(df)
    > df = df.rename(columns = strip_column_names(df.columns))

    ## observe, by comparison
    > df2 = df.rename(columns = strip_column_names(df.columns,
        keep_paren_contents=False))
    """

    # strip/replace punctuation
    new_cols = [
        _strip_column_name(col, keep_paren_contents=keep_paren_contents)
        for col in cols]

    if len(new_cols) != len(set(new_cols)):
        warn_str = 'Warning: strip_column_names (if run) would introduce duplicate names.'
        warn_str += ' Reverting column names to the original.'

        warnings.warn(warn_str, Warning)
        print('Warning: strip_column_names would introduce duplicate names. Please fix & try again.')
        return dict(zip(cols, cols))

    return dict(zip(cols, new_cols))



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from nose.tools import eq_

from cohorts.rounding import float_str

ROUNDING_CASES = {2: '2.00',
                  2.0: '2.00',
                  2.011: '2.01',
                  4.015: '4.02',
                  4.025: '4.03',
                  3.015: '3.02',
                  3.025: '3.03',
                  0.001: '0.001',
                  0.0010: '0.001',
                  0.00104: '0.0010',
                  0.00109: '0.0011',
                  0.00111: '0.0011',
                  4.5: '4.50',
                  1235345.132424: '1235345.13',
                  0.009: '0.009',
                  234.009: '234.01',
                  3.000001: '3.00',
                  13212324.0: '13212324.00',
                  -1.1: '-1.10'}

def test_rounding():
    for f, expected_str in ROUNDING_CASES.items():
        eq_(expected_str, float_str(f))



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from cohorts import Cohort, DataFrameLoader

import pandas as pd
from nose.tools import eq_

from .test_basic import make_simple_cohort

def test_df_loading():
    cohort = make_simple_cohort()
    df_hello = pd.DataFrame({"the_id": ["1", "5", "7"],
                             "hello_value": ["hello", "goodbye", "hello"]})
    def load_df():
        return df_hello
    df_loader = DataFrameLoader("hello", load_df, join_on="the_id")
    cohort.df_loaders = [df_loader]

    df = cohort.as_dataframe(join_with="hello")
    eq_(len(df), 2)
    # pylint: disable=no-member
    # pylint gets confused by as_dataframe's return type
    eq_(set(df.patient_id), set(["1", "5"]))



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import warnings

def _provenance_str(provenance):
    """Utility function used by compare_provenance to print diff
    """
    return ["%s==%s" % (key, value) for (key, value) in provenance]

def compare_provenance(
        this_provenance, other_provenance,
        left_outer_diff = "In current but not comparison",
        right_outer_diff = "In comparison but not current"):
    """Utility function to compare two abritrary provenance dicts
    returns number of discrepancies.

    Parameters
    ----------
    this_provenance: provenance dict (to be compared to "other_provenance")
    other_provenance: comparison provenance dict

    (optional)
    left_outer_diff: description/prefix used when printing items in this_provenance but not in other_provenance
    right_outer_diff: description/prefix used when printing items in other_provenance but not in this_provenance

    Returns
    -----------
    Number of discrepancies (0: None)
    """
    ## if either this or other items is null, return 0
    if (not this_provenance or not other_provenance):
        return 0

    this_items = set(this_provenance.items())
    other_items = set(other_provenance.items())

    # Two-way diff: are any modules introduced, and are any modules lost?
    new_diff = this_items.difference(other_items)
    old_diff = other_items.difference(this_items)
    warn_str = ""
    if len(new_diff) > 0:
        warn_str += "%s: %s" % (
            left_outer_diff,
            _provenance_str(new_diff))
    if len(old_diff) > 0:
        warn_str += "%s: %s" % (
            right_outer_diff,
            _provenance_str(old_diff))

    if len(warn_str) > 0:
        warnings.warn(warn_str, Warning)

    return(len(new_diff)+len(old_diff))



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Collection(object):
    def __init__(self, elements):
        self.elements = elements

    def short_string(self):
        """
        Compact string representation which doesn't print any of the
        collection elements.
        """
        file_str = ""
        return "<%s with %d elements>" % (
            self.__class__.__name__,
            len(self))

    def to_string(self, limit=None):
        """
        Create a string representation of this collection, showing up to
        `limit` items.
        """
        header = self.short_string()
        if len(self) == 0:
            return header
        contents = ""
        element_lines = [
            "  -- %s" % (element,)
            for element in self.elements[:limit]
        ]
        contents = "\n".join(element_lines)

        if limit is not None and len(self.elements) > limit:
            contents += "\n  ... and %d more" % (len(self) - limit)
        return "%s\n%s" % (header, contents)

    def __str__(self):
        return self.to_string(limit=50)

    def __repr__(self):
        return str(self)

    def __len__(self):
        return len(self.elements)

    def __iter__(self):
        return iter(self.elements)

    def __getitem__(self, idx):
        return self.elements[idx]



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import os
from os import path
from codecs import open
from setuptools import setup
import versioneer

current_directory = os.path.dirname(__file__)
readme_filename = "README.md"
readme_path = os.path.join(current_directory, readme_filename)

readme = ""
try:
    with open(readme_path, "r") as f:
        readme = f.read()
except IOError as e:
    print(e)
    print("Failed to open %s" % readme_path)

try:
    import pypandoc
    readme = pypandoc.convert(readme, to="rst", format="md")
except ImportError as e:
    print(e)
    print("Failed to convert %s to reStructuredText", readme_filename)
    pass

# get the dependencies and installs
with open(path.join(current_directory, "requirements.txt"), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [req.strip() for req in all_reqs if 'git+' not in req]
dependency_links = [req.strip().replace('git+','') for req in all_reqs if 'git+' in req]

if __name__ == "__main__":
    setup(
        name="cohorts",
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
        description="Utilities for analyzing mutations and neoepitopes in patient cohorts",
        author="Tavi Nathanson",
        author_email="tavi {dot} nathanson {at} gmail {dot} com",
        url="https://github.com/tavinathanson/cohorts",
        license="http://www.apache.org/licenses/LICENSE-2.0.html",
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Environment :: Console",
            "Operating System :: OS Independent",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: Python",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
        ],
        install_requires=install_requires,
        dependency_links=dependency_links,
        long_description=readme,
        packages=["cohorts"],
    )


# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

class DataFrameLoader(object):
    """
    Wraps a `DataFrame` with some information on how to join it.

    Parameters
    __________
    name : str
        The name of the dataframe, to easily reference it.
    load_dataframe : function
        A function that returns the `DataFrame` object.
    join_on : str
        The column of the `DataFrame` to join on (i.e. the patient
        ID column name).
    """
    def __init__(self,
                 name,
                 load_dataframe,
                 join_on="patient_id"):
        self.name = name
        self.load_dataframe = load_dataframe
        self.join_on = join_on



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from .variant_filters import no_filter, effect_expressed_filter
from .utils import first_not_none_param

import numpy as np
from varcode.effects import Substitution
from varcode.common import memoize

def snv_count(row, cohort, filter_fn=None,
              normalized_per_mb=None, **kwargs):
    filter_fn = first_not_none_param([filter_fn, cohort.filter_fn], no_filter)
    normalized_per_mb = first_not_none_param([normalized_per_mb, cohort.normalized_per_mb], False)
    patient_id = row["patient_id"]
    patient_variants = cohort.load_variants(
        patients=[cohort.patient_from_id(patient_id)],
        filter_fn=filter_fn,
        **kwargs)
    if patient_id in patient_variants:
        count = len(patient_variants[patient_id])
        if normalized_per_mb:
            count /= float(get_patient_to_mb(cohort)[patient_id])
        return count
    return np.nan

def nonsynonymous_snv_count(row, cohort, filter_fn=None,
                            normalized_per_mb=None, **kwargs):
    filter_fn = first_not_none_param([filter_fn, cohort.filter_fn], no_filter)
    normalized_per_mb = first_not_none_param([normalized_per_mb, cohort.normalized_per_mb], False)
    patient_id = row["patient_id"]
    # This only loads one effect per variant.
    patient_nonsynonymous_effects = cohort.load_effects(
        only_nonsynonymous=True,
        patients=[cohort.patient_from_id(patient_id)],
        filter_fn=filter_fn,
        **kwargs)
    if patient_id in patient_nonsynonymous_effects:
        count = len(patient_nonsynonymous_effects[patient_id])
        if normalized_per_mb:
            count /= float(get_patient_to_mb(cohort)[patient_id])
        return count
    return np.nan

def missense_snv_count(row, cohort, filter_fn=None,
                       normalized_per_mb=None, **kwargs):
    filter_fn = first_not_none_param([filter_fn, cohort.filter_fn], no_filter)
    normalized_per_mb = first_not_none_param([normalized_per_mb, cohort.normalized_per_mb], False)
    patient_id = row["patient_id"]
    def missense_filter_fn(filterable_effect):
        assert filter_fn is not None, "filter_fn should never be None, but it is."
        return (type(filterable_effect.effect) == Substitution and
                filter_fn(filterable_effect))
    # This only loads one effect per variant.
    patient_missense_effects = cohort.load_effects(
        only_nonsynonymous=True,
        patients=[cohort.patient_from_id(patient_id)],
        filter_fn=missense_filter_fn,
        **kwargs)
    if patient_id in patient_missense_effects:
        count = len(patient_missense_effects[patient_id])
        if normalized_per_mb:
            count /= float(get_patient_to_mb(cohort)[patient_id])
        return count
    return np.nan

def neoantigen_count(row, cohort, filter_fn=None,
                     normalized_per_mb=None, **kwargs):
    filter_fn = first_not_none_param([filter_fn, cohort.filter_fn], no_filter)
    normalized_per_mb = first_not_none_param([normalized_per_mb, cohort.normalized_per_mb], False)
    patient_id = row["patient_id"]
    patient = cohort.patient_from_id(row["patient_id"])
    patient_neoantigens = cohort.load_neoantigens(patients=[patient],
                                                  filter_fn=filter_fn,
                                                  **kwargs)
    if patient_id in patient_neoantigens:
        patient_neoantigens_df = patient_neoantigens[patient_id]
        count = len(patient_neoantigens_df)
        if normalized_per_mb:
            count /= float(get_patient_to_mb(cohort)[patient_id])
        return count
    return np.nan

def expressed_missense_snv_count(row, cohort, filter_fn=None,
                                 normalized_per_mb=None):
    filter_fn = first_not_none_param([filter_fn, cohort.filter_fn], no_filter)
    normalized_per_mb = first_not_none_param([normalized_per_mb, cohort.normalized_per_mb], False)
    def expressed_filter_fn(filterable_effect):
        assert filter_fn is not None, "filter_fn should never be None, but it is."
        return filter_fn(filterable_effect) and effect_expressed_filter(filterable_effect)
    return missense_snv_count(row, cohort, filter_fn=expressed_filter_fn,
                              normalized_per_mb=normalized_per_mb)

def expressed_neoantigen_count(row, cohort, **kwargs):
    return neoantigen_count(row, cohort, only_expressed=True, **kwargs)

@memoize
def get_patient_to_mb(cohort):
    patient_to_mb = dict(cohort.as_dataframe(join_with="ensembl_coverage")[["patient_id", "MB"]].to_dict("split")["data"])
    return patient_to_mb
